import copy
import keyword as _keyword
import re as _re
from collections.abc import MutableMapping, MutableSequence
from itertools import chain
from numbers import Number

import numpy

from .model.linear import LinearComponent, LinearFunction
from .util.naming import parenthize
from .util.touch_notifier import TouchNotify

_chi = "\u03C7"
_beta = "\u03B2"
_hahook = "\u04FC"

_ParameterRef_repr_txt = "P"
_DataRef_repr_txt = "X"


class Role(type):
    def __getattr__(cls, key):
        try:
            return super().getattr(key)
        except AttributeError:
            if cls is ParameterRef or cls is DataRef or cls is CombinedRef:
                return cls(key)
            else:
                raise


_data_description_catalog = {}


class DataRef(str, metaclass=Role):
    def __new__(cls, name, **kwarg):
        if hasattr(str, name):
            raise NameError(
                f"cannot create DataRef with the name of a str method ({name})"
            )
        if name in ("_descrip", "descrip", "_role"):
            raise NameError(f"cannot create DataRef with the name ({name})")
        try:
            raise TypeError("debugging")  # The getattr here is not really needed?
            return getattr(cls, name)
        except Exception:
            self = super().__new__(cls, name)
            self._role = "data"
            return self

    def __getnewargs__(self):
        return (str(self),)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)

    def __init__(self, name, descrip=None):
        if descrip is not None:
            _data_description_catalog[self] = descrip
        super().__init__()

    @property
    def descrip(self):
        if self in _data_description_catalog:
            return _data_description_catalog[self]
        return None

    @descrip.setter
    def descrip(self, value):
        if value is not None:
            _data_description_catalog[self] = value

    def __repr__(self):
        s = super().__str__()
        if _re.match("[_A-Za-z][_a-zA-Z0-9]*$", s) and not _keyword.iskeyword(s):
            return f"{_DataRef_repr_txt}.{s}"
        else:
            return f"{_DataRef_repr_txt}('{s}')"

    def __add__(self, other):
        if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
            return LinearComponent(data=str(self), param="CONSTANT") + other
        return DataRef(f"{parenthize(self)}+{parenthize(other, True)}")

    def __radd__(self, other):
        if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
            return other + LinearComponent(data=str(self), param="CONSTANT")
        return DataRef(f"{parenthize(other)}+{parenthize(self, True)}")

    def __sub__(self, other):
        if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
            return LinearComponent(data=str(self), param="CONSTANT") - other
        return DataRef(f"{parenthize(self)}-{parenthize(other, True)}")

    def __rsub__(self, other):
        if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
            return other - LinearComponent(data=str(self), param="CONSTANT")
        return DataRef(f"{parenthize(other)}-{parenthize(self, True)}")

    def __mul__(self, other):
        if isinstance(other, LinearFunction):
            trial = LinearFunction()
            for component in other:
                trial += self * component
            return trial
        if isinstance(other, LinearComponent):
            return LinearComponent(data=str(self * other.data), param=str(other.param))
        if type(other) is ParameterRef:
            return LinearComponent(data=str(self), param=str(other))
        if type(other) is _param_multiply:
            if isinstance(other._left, ParameterRef):
                return (
                    LinearComponent(data=str(self), param=str(other._left))
                    * other._right
                )
            else:
                return (
                    LinearComponent(data=str(self), param=str(other._right))
                    * other._left
                )
        if type(other) is _param_negate:
            return LinearComponent(data=str(-self), param=str(other._orig))
        if other == 0:
            return 0
        return DataRef(f"{parenthize(self)}*{parenthize(other, True)}")

    def __rmul__(self, other):
        if isinstance(other, LinearFunction):
            raise NotImplementedError
        if isinstance(other, LinearComponent):
            return LinearComponent(data=str(other.data * self), param=str(other.param))
        if type(other) is ParameterRef:
            return LinearComponent(data=str(self), param=str(other))
        if type(other) is _param_negate:
            return LinearComponent(data=str(-self), param=str(other._orig))
        if other == 0:
            return 0
        return DataRef(f"{parenthize(other)}*{parenthize(self, True)}")

    def __truediv__(self, other):
        if isinstance(other, LinearFunction):
            raise NotImplementedError
        if isinstance(other, LinearComponent):
            raise NotImplementedError
        if isinstance(other, ParameterRef):
            raise NotImplementedError
        return DataRef(f"{parenthize(self)}/{parenthize(other, True)}")

    def __rtruediv__(self, other):
        if isinstance(other, LinearFunction):
            raise NotImplementedError
        if isinstance(other, LinearComponent):
            raise NotImplementedError
        if isinstance(other, ParameterRef):
            raise NotImplementedError
        return DataRef(f"{parenthize(other)}/{parenthize(self, True)}")

    def __and__(self, other):
        if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
            raise NotImplementedError
        return DataRef(f"{parenthize(self)}&{parenthize(other, True)}")

    def __rand__(self, other):
        if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
            raise NotImplementedError
        return DataRef(f"{parenthize(other)}&{parenthize(self, True)}")

    def __or__(self, other):
        if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
            raise NotImplementedError
        return DataRef(f"{parenthize(self)}|{parenthize(other, True)}")

    def __ror__(self, other):
        if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
            raise NotImplementedError
        return DataRef(f"{parenthize(other)}|{parenthize(self, True)}")

    def __xor__(self, other):
        if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
            raise NotImplementedError
        return DataRef(f"{parenthize(self)}^{parenthize(other, True)}")

    def __rxor__(self, other):
        if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
            raise NotImplementedError
        return DataRef(f"{parenthize(other)}^{parenthize(self, True)}")

    def __neg__(self):
        return DataRef(f"-({self})")

    def __pos__(self):
        return self

    def __eq__(self, other):
        if isinstance(other, DataRef):
            if repr(other) == repr(self):
                return True
        if isinstance(other, ParameterRef):
            return False
        if isinstance(other, str):
            if other == super().__str__():
                return True
        return False

    def __hash__(self):
        return hash(super().__str__())

    def eval(self, namespace=None, *, globals=None, **more_namespace):
        import numpy

        from .util.common_functions import hard_sigmoid, piece

        use_namespace = {
            "exp": numpy.exp,
            "log": numpy.log,
            "log1p": numpy.log1p,
            "fabs": numpy.fabs,
            "sqrt": numpy.sqrt,
            "absolute": numpy.absolute,
            "isnan": numpy.isnan,
            "isfinite": numpy.isfinite,
            "logaddexp": numpy.logaddexp,
            "fmin": numpy.fmin,
            "fmax": numpy.fmax,
            "nan_to_num": numpy.nan_to_num,
            "piece": piece,
            "hard_sigmoid": hard_sigmoid,
        }
        if namespace is not None:
            use_namespace.update(namespace)
        use_namespace.update(more_namespace)
        return eval(self, globals, use_namespace)

    def __getattr__(self, item):
        return DataRef(".".join([self, item]))

    def __copy__(self):
        # str are immutable
        return self

    def __deepcopy__(self, memodict):
        return copy.deepcopy(str(self), memodict)


class ParameterRef(str, metaclass=Role):
    """
    An abstract reference to a parameter, which may or may not be in any given model.

    Parameters
    ----------
    name : str
            The name of the parameter to reference.
    default : numeric or None
            When a targeted model does not include the referenced parameter, use
            this value for :meth:`value` and :meth:`str`. If given as None, those
            methods will raise an exception.
    fmt : str or None
            The format string to use for :meth:`str`.
    """

    def __new__(
        cls, name, default=None, fmt=None, default_value=None, format=None, scale=None
    ):
        if fmt is None and format is not None:
            fmt = format
        if fmt is None:
            fmt = "{}"
        if default is None and default_value is not None:
            default = default_value
        self = super().__new__(cls, name)
        self._fmt = fmt
        self._name = name
        self._role = "parameter"
        self._default_value = default
        return self

    def __repr__(self):
        s = super().__str__()
        if _re.match("[_A-Za-z][_a-zA-Z0-9]*$", s) and not _keyword.iskeyword(s):
            return f"{_ParameterRef_repr_txt}.{s}"
        else:
            return f"{_ParameterRef_repr_txt}('{s}')"

    def __str__(self):
        return super().__str__()

    def __xml__(self, resolve_parameters=None):
        from xmle import Elem

        x = Elem("div")
        a_p = x.elem("div", attrib={"class": "tooltipped"}, text=repr(self))
        if resolve_parameters is not None:
            plabel = "{:.3g}".format(
                resolve_parameters.pvalue(self, default_value="This is a Parameter")
            )
        else:
            plabel = "This is a Parameter"
        a_p.elem("span", attrib={"class": "tooltiptext"}, text=plabel)
        return x

    def value(self, m):
        """
        Get the value of the parameter in a given model.

        Parameters
        ----------
        m : Model
                The model from which to extract a parameter value.

        Raises
        ------
        LarchError
                When the model does not contain a parameter with the same
                name as this ParameterRef, and the default_value for this
                ParameterRef is None.
        """
        try:
            v = m.pvalue(str(self))
        except KeyError:
            if self._default_value is not None:
                v = self._default_value
            else:
                raise
        return v

    def name(self, n):
        """
        Set or re-set the name of the parameter referenced.

        Parameters
        ----------
        n : str
                The new name of the parameter.

        Returns
        -------
        ParameterRef
                This method returns the self object, to facilitate method chaining.
        """
        self._name = n
        return self

    def fmt(self, format):
        """
        Set or re-set the format string for the parameter referenced.

        Parameters
        ----------
        format : str
            The new format string. This will be used to format the parameter
            value when calling :meth:`str`.

        Returns
        -------
        ParameterRef
                This method returns the self object, to facilitate method chaining.
        """
        if format is None or format == "":
            self._fmt = "{}"
        else:
            self._fmt = format
        try:
            self._fmt.format(0)
        except Exception as err:
            raise ValueError("invalid formating string") from err
        return self

    def default_value(self, val):
        """
        Set or re-set the default value.

        Parameters
        ----------
        val : numeric or None
                The new default value. This will be used by :meth:`value` when
                the parameter is not included in the targeted model. Set to None
                to raise an exception instead.

        Returns
        -------
        ParameterRef
                This method returns the self object, to facilitate method chaining.
        """
        self._default_value = val
        return self

    def strf(self, m, fmt=None):
        """
        Get the :meth:`value` of the parameter in a given model as a string.

        The string is formated using the :meth:`fmt` string if given. If not
        given, Python's default string formatting is used.

        Parameters
        ----------
        m : Model
                The model from which to extract a parameter value.
        fmt : str or None
                If not None, a format string which will be used, overriding any
                previous setting by a :meth:`fmt` command.  The override is
                valid for this method call only, and does not change the
                format for future calls.

        Raises
        ------
        KeyError
                When the model does not contain a parameter with the same
                name as this ParameterRef, and the default_value for this
                ParameterRef is None.
        """
        try:
            if fmt is not None:
                return fmt.format(self.value(m))
            else:
                return self._fmt.format(self.value(m))
        except KeyError:
            return "NA"

    def getname(self):
        return self._name

    def valid(self, m):
        """
        Check if this ParameterRef would give a value for a given model.

        Parameters
        ----------
        m : Model
                The model from which to extract a parameter value.

        Returns
        -------
        bool
                False if the value method would raise an exception, and True otherwise.
        """
        if str(self) in m:
            return True
        if self._default_value is not None:
            return True
        return False

    def __add__(self, other):
        if other == 0:
            return self
        if isinstance(other, (DataRef, LinearComponent, LinearFunction)):
            return LinearComponent(param=str(self), data="1") + other
        return _param_add(self, other)

    def __radd__(self, other):
        if other == 0:
            return self
        if isinstance(other, (DataRef, LinearComponent, LinearFunction)):
            return other + LinearComponent(param=str(self), data="1")
        return _param_add(other, self)

    def __sub__(self, other):
        if other == 0:
            return self
        if isinstance(other, (DataRef, LinearComponent, LinearFunction)):
            return LinearComponent(param=str(self), data="1") - other
        return _param_subtract(self, other)

    def __rsub__(self, other):
        if other == 0:
            return self
        if isinstance(other, (DataRef, LinearComponent, LinearFunction)):
            return other - LinearComponent(param=str(self), data="1")
        return _param_subtract(other, self)

    def __mul__(self, other):
        if isinstance(other, (LinearComponent, LinearFunction)):
            raise TypeError("cannot do ParameterRef * LinearComponent")
        if type(other) is DataRef:
            if isinstance(self, _param_multiply):
                if isinstance(self._left, ParameterRef):
                    return (
                        LinearComponent(data=str(other), param=str(self._left))
                        * self._right
                    )
                else:
                    return (
                        LinearComponent(data=str(other), param=str(self._right))
                        * self._left
                    )
            return LinearComponent(data=str(other), param=str(self))
        return _param_multiply(self, other)

    def __rmul__(self, other):
        if isinstance(other, (LinearComponent, LinearFunction)):
            raise TypeError("cannot do LinearComponent * ParameterRef")
        if type(other) is DataRef:
            if isinstance(self, _param_multiply):
                if isinstance(self._left, ParameterRef):
                    return (
                        LinearComponent(data=str(other), param=str(self._left))
                        * self._right
                    )
                else:
                    return (
                        LinearComponent(data=str(other), param=str(self._right))
                        * self._left
                    )
            return LinearComponent(data=str(other), param=str(self))
        return _param_multiply(other, self)

    def __truediv__(self, other):
        if isinstance(other, DataRef):
            return self * (1 / other)
        if isinstance(other, (LinearComponent, LinearFunction)):
            raise TypeError("cannot do ParameterRef / LinearComponent")
        return _param_divide(self, other)

    def __rtruediv__(self, other):
        if isinstance(other, DataRef):
            raise TypeError("cannot do DataRef / ParameterRef")
        if isinstance(other, (LinearComponent, LinearFunction)):
            raise TypeError("cannot do LinearComponent / ParameterRef")
        return _param_divide(other, self)

    def __neg__(self):
        return _param_negate(self)

    def __pos__(self):
        return self

    def __eq__(self, other):
        if isinstance(other, ParameterRef):
            if repr(other) == repr(self):
                return True
        if isinstance(other, DataRef):
            return False
        if isinstance(other, str):
            if other == super().__str__():
                return True
        return False

    def __hash__(self):
        return hash(repr(self))

    def _to_Linear(self):
        return LinearComponent(param=str(self), data="1")

    def _is_scaled_parameter(self):
        if type(self) is ParameterRef:
            return (self, 1.0)
        return None

    @property
    def unscaled_parameter(self):
        y = self._is_scaled_parameter()
        if y is None:
            return None
        return y[0]

    @property
    def scaling_of_parameter(self):
        y = self._is_scaled_parameter()
        if y is None:
            return 1
        return y[1]

    def __le__(self, other):
        from .model.jaxmodel.guidance import ParameterScaledOrdering

        try:
            oth = other._is_scaled_parameter()
        except AttributeError:
            raise TypeError("other is not scaled parameter") from None
        if oth:
            return ParameterScaledOrdering(None, *self._is_scaled_parameter(), *oth)
        else:
            raise TypeError("other is not scaled parameter")

    def __ge__(self, other):
        from .jaxmodel.guidance import ParameterScaledOrdering

        try:
            oth = other._is_scaled_parameter()
        except AttributeError:
            raise TypeError("other is not scaled parameter") from None
        if oth:
            return ParameterScaledOrdering(
                None, *oth, *self._is_scaled_parameter(), snap_out="greater?"
            )
        else:
            raise TypeError("other is not scaled parameter")

    def lessthan(self, other):
        s = self._is_scaled_parameter()
        if s is None:
            raise TypeError("self is not scaled parameter")
        from .jaxmodel.guidance import ParameterBound, ParameterScaledOrdering

        try:
            oth = other._is_scaled_parameter()
        except AttributeError:
            if isinstance(other, Number):
                return ParameterBound(None, *self._is_scaled_parameter(), None, other)
            raise TypeError("other is not scaled parameter") from None
        if oth:
            return ParameterScaledOrdering(None, *self._is_scaled_parameter(), *oth)
        else:
            raise TypeError("other is not scaled parameter")

    def __lshift__(self, other):
        return self.lessthan(other)

    def greaterthan(self, other):
        s = self._is_scaled_parameter()
        if s is None:
            raise TypeError("self is not scaled parameter")
        from .jaxmodel.guidance import ParameterBound, ParameterScaledOrdering

        try:
            oth = other._is_scaled_parameter()
        except AttributeError:
            if isinstance(other, Number):
                return ParameterBound(None, None, other, *self._is_scaled_parameter())
            raise TypeError("other is not scaled parameter") from None
        if oth:
            return ParameterScaledOrdering(
                None, *oth, *self._is_scaled_parameter(), snap_out="greater"
            )
        else:
            raise TypeError("other is not scaled parameter")

    def __rshift__(self, other):
        return self.greaterthan(other)

    def equalto(self, other):
        s = self._is_scaled_parameter()
        if s is None:
            raise TypeError("self is not scaled parameter")
        from .jaxmodel.guidance import ParameterScaledEquality

        try:
            oth = other._is_scaled_parameter()
        except AttributeError:
            raise TypeError("other is not scaled parameter") from None
        if oth:
            return ParameterScaledEquality(None, *self._is_scaled_parameter(), *oth)
        else:
            raise TypeError("other is not scaled parameter")

    def __matmul__(self, other):
        return self.equalto(other)

    def __rmatmul__(self, other):
        try:
            return other.equalto(self)
        except AttributeError:
            raise NotImplementedError from None

    def between(self, lower, upper):
        return self.greaterthan(lower), self.lessthan(upper)

    def __copy__(self):
        # str are immutable
        return self

    def __deepcopy__(self, memodict):
        return copy.deepcopy(str(self), memodict)


class _param_math_binaryop(ParameterRef):
    def __new__(cls, left, right):
        self = super().__new__(cls, "")
        self._left = left
        self._right = right
        if isinstance(self._left, ParameterRef) and isinstance(
            self._right, ParameterRef
        ):
            if left._fmt == "{}" and right._fmt != "{}":
                self._fmt = right._fmt
            else:
                self._fmt = left._fmt
        elif isinstance(self._left, ParameterRef):
            self._fmt = left._fmt
        elif isinstance(self._right, ParameterRef):
            self._fmt = right._fmt
        else:
            self._fmt = "{}"
        self._name = ""
        self._role = "parameter_math"
        return self

    def __getnewargs__(self):
        return (self._left, self._right)


class _param_add(_param_math_binaryop):
    def value(self, m):
        if isinstance(self._left, ParameterRef):
            x = self._left.value(m)
        else:
            x = self._left
        if isinstance(self._right, ParameterRef):
            x += self._right.value(m)
        else:
            x += self._right
        return x

    def valid(self, m):
        if isinstance(self._left, ParameterRef):
            x = self._left.valid(m)
        else:
            x = True
        if isinstance(self._right, ParameterRef):
            x &= self._right.valid(m)
        return x

    def __repr__(self):
        return f"({repr(self._left)} + {repr(self._right)})"

    def __str__(self):
        return f"{repr(self._left)} + {repr(self._right)}"

    def _to_Linear(self):
        return self._left._to_Linear() + self._right._to_Linear()

    def __add__(self, other):
        if isinstance(other, (DataRef, LinearComponent, LinearFunction)):
            return self._to_Linear() + other
        return super().__add__(other)


class _param_subtract(_param_math_binaryop):
    def value(self, m):
        if isinstance(self._left, ParameterRef):
            x = self._left.value(m)
        else:
            x = self._left
        if isinstance(self._right, ParameterRef):
            x -= self._right.value(m)
        else:
            x -= self._right
        return x

    def valid(self, m):
        if isinstance(self._left, ParameterRef):
            x = self._left.valid(m)
        else:
            x = True
        if isinstance(self._right, ParameterRef):
            x &= self._right.valid(m)
        return x

    def __repr__(self):
        return f"({repr(self._left)} - {repr(self._right)})"

    def __str__(self):
        return f"{repr(self._left)} - {repr(self._right)}"


class _param_multiply(_param_math_binaryop):
    def value(self, m):
        if isinstance(self._left, ParameterRef):
            x = self._left.value(m)
        else:
            x = self._left
        if isinstance(self._right, ParameterRef):
            x *= self._right.value(m)
        else:
            x *= self._right
        return x

    def valid(self, m):
        if isinstance(self._left, ParameterRef):
            x = self._left.valid(m)
        else:
            x = True
        if isinstance(self._right, ParameterRef):
            x &= self._right.valid(m)
        return x

    def __repr__(self):
        return f"({repr(self._left)} * {repr(self._right)})"

    def __str__(self):
        return f"{repr(self._left)} * {repr(self._right)}"

    def _is_scaled_parameter(self):
        if type(self._left) is ParameterRef and isinstance(self._right, Number):
            return (self._left, self._right)
        elif type(self._right) is ParameterRef and isinstance(self._left, Number):
            return (self._right, self._left)
        else:
            return None

    def __mul__(self, other):
        x = self._is_scaled_parameter()
        if x and isinstance(other, Number):
            return x[0] * (x[1] * other)
        return super().__mul__(other)

    def __rmul__(self, other):
        x = self._is_scaled_parameter()
        if x and isinstance(other, Number):
            return x[0] * (x[1] * other)
        return super().__mul__(other)

    def __eq__(self, other):
        if isinstance(other, _param_multiply):
            if (self._left == other._left) and (self._right == other._right):
                return True
            elif (self._left == other._right) and (self._right == other._left):
                return True
            else:
                return False
        elif isinstance(other, ParameterRef):
            if (self._left == other) and (self._right == 1):
                return True
            elif (self._left == 1) and (self._right == other):
                return True
            else:
                return False
        else:
            raise NotImplementedError(f"type of other is {type(other)}")

    def __ne__(self, other):
        return not (self == other)


class _param_divide(_param_math_binaryop):
    def value(self, m):
        if isinstance(self._left, ParameterRef):
            x = self._left.value(m)
        else:
            x = self._left
        try:
            if isinstance(self._right, ParameterRef):
                x /= self._right.value(m)
            else:
                x /= self._right
        except ZeroDivisionError:
            return float("NaN")
        return x

    def valid(self, m):
        if isinstance(self._left, ParameterRef):
            x = self._left.valid(m)
        else:
            x = True
        if isinstance(self._right, ParameterRef):
            x &= self._right.valid(m)
        return x

    def __repr__(self):
        return f"({repr(self._left)} / {repr(self._right)})"

    def __str__(self):
        return f"{repr(self._left)} / {repr(self._right)}"


class _param_negate(ParameterRef):
    def __new__(cls, orig):
        self = super().__new__(cls, "")
        self._orig = orig
        self._fmt = orig._fmt
        self._name = f"-({orig.getname()})"
        self._role = "parameter_math"
        return self

    def value(self, m):
        if isinstance(self._orig, ParameterRef):
            return -self._orig.value(m)
        else:
            return -self._orig

    def valid(self, m):
        if isinstance(self._orig, ParameterRef):
            return self._orig.valid(m)
        else:
            return True

    def __repr__(self):
        return f"-({repr(self._orig)})"

    def __str__(self):
        return f"-({repr(self._orig)})"


try:
    from numpy import exp as _exp
    from numpy import log as _log
except ImportError:
    from math import exp as _exp
    from math import log as _log


def log(x):
    if isinstance(x, DataRef):
        return DataRef(f"log({x})")
    else:
        return _log(x)


def exp(x):
    if isinstance(x, DataRef):
        return DataRef(f"exp({x})")
    else:
        return _exp(x)


class CombinedRef(metaclass=Role):
    def __new__(cls, s):
        return LinearComponent(data=str(s), param=str(s))


# +++


class LinearComponent2(tuple):
    __slots__ = []

    def __new__(cls, param, data="1", scale=1):
        return tuple.__new__(cls, (str(param), str(data), scale))

    @property
    def param(self):
        return ParameterRef(self[0])

    @property
    def data(self):
        return DataRef(self[1])

    @property
    def scale(self):
        return self[2]

    def __pos__(self):
        return self

    def __repr__(self):
        try:
            if self.scale == 1.0:
                try:
                    data_is_1 = float(self.data) == 1
                except Exception:
                    data_is_1 = False
                if data_is_1:
                    return f"({self.param!r})"
                else:
                    return f"({self.param!r} * {self.data!r})"
            return f"({self.param!r} * {self.scale} * {self.data!r})"
        except AttributeError:
            return f"<{self.__class__.__name__} {id(self)} with error>"

    def __add__(self, other):
        if other == ():
            return self
        elif isinstance(other, LinearComponent2):
            return LinearFunction2([self, other])
        elif isinstance(other, LinearFunction2):
            return LinearFunction2([self, *other])
        elif other == 0:
            return self
        elif isinstance(other, ParameterRef):
            return LinearFunction2([self, LinearComponent2(param=other)])
        else:
            raise NotImplementedError()

    def __radd__(self, other):
        if other == 0:
            return self
        return other.__add__(self)

    def __mul__(self, other):
        if isinstance(
            other,
            (
                int,
                float,
            ),
        ):
            return self.__class__(
                param=self.param,
                data=self.data,
                scale=self.scale * other,
            )
        elif isinstance(other, (DataRef,)):
            return self.__class__(
                param=self.param,
                data=self.data * other,
                scale=self.scale,
            )
        else:
            raise TypeError(
                f"unsupported operand type(s) for {self.__class__.__name__}: {type(other)}"
            )

    def __rmul__(self, other):
        if isinstance(
            other,
            (
                int,
                float,
            ),
        ):
            return self.__class__(
                param=self.param,
                data=self.data,
                scale=self.scale * other,
            )
        elif isinstance(other, (DataRef,)):
            return self.__class__(
                param=self.param,
                data=self.data * other,
                scale=self.scale,
            )
        else:
            raise TypeError(
                f"unsupported operand type(s) for "
                f"{self.__class__.__name__}: {type(other)}"
            )

    def __iter__(self):
        return iter(LinearFunction2([self]))

    def __eq__(self, other):
        if isinstance(other, LinearFunction2) and len(other) == 1:
            other = other[0]
        if not isinstance(other, LinearComponent2):
            return False
        if self.param != other.param:
            return False
        if self.data != other.data:
            return False
        if self.scale != other.scale:
            return False
        return True

    def __xml__(self, exponentiate_parameter=False, resolve_parameters=None):
        from xmle import Elem

        x = Elem("div")
        # x << tooltipped_style()

        if resolve_parameters is not None:
            if exponentiate_parameter:
                plabel = resolve_parameters.pvalue(
                    self.param, default_value="This is a Parameter"
                )
                if not isinstance(plabel, str):
                    plabel = f"exp({plabel:.3g})={numpy.exp(plabel):.3g}"
            else:
                plabel = "{:.3g}".format(
                    resolve_parameters.pvalue(
                        self.param, default_value="This is a Parameter"
                    )
                )
        else:
            plabel = "This is a Parameter"

        data_tail = " * "
        try:
            if float(self.data) == 1:
                data_tail = ""
        except Exception:
            pass

        if self.scale == 1.0:
            if exponentiate_parameter:
                x.elem("span", tail="exp(")
                a_p = x.elem(
                    "div",
                    attrib={"class": "tooltipped"},
                    text=repr(self.param),
                    tail=")" + data_tail,
                )
                a_p.elem("span", attrib={"class": "tooltiptext"}, text=plabel)
            else:
                a_p = x.elem(
                    "div",
                    attrib={"class": "tooltipped"},
                    text=repr(self.param),
                    tail=data_tail,
                )
                a_p.elem("span", attrib={"class": "tooltiptext"}, text=plabel)
        else:
            if exponentiate_parameter:
                x.elem("span", tail="exp(")
                a_p = x.elem(
                    "div",
                    attrib={"class": "tooltipped"},
                    text=repr(self.param),
                    tail=f" * {self.scale}){data_tail}",
                )
                a_p.elem("span", attrib={"class": "tooltiptext"}, text=plabel)
            else:
                a_p = x.elem(
                    "div",
                    attrib={"class": "tooltipped"},
                    text=repr(self.param),
                    tail=f" * {self.scale}{data_tail}",
                )
                a_p.elem("span", attrib={"class": "tooltiptext"}, text=plabel)
        if data_tail == " * ":
            a_x = x.elem("div", attrib={"class": "tooltipped"}, text=repr(self.data))
            a_x.elem("span", attrib={"class": "tooltiptext"}, text="This is Data")
        return x

    def _repr_html_(self):
        return self.__xml__().tostring()

    def _evaluate(self, p_getter, x_namespace=None, **kwargs):
        return (
            self.scale
            * p_getter(str(self.param))
            * self.data.eval(namespace=x_namespace, **kwargs)
        )

    def __copy__(self):
        return self.__class__(
            param=self.param,
            data=self.data,
            scale=self.scale,
        )


def _try_mangle(instance):
    try:
        instance.mangle()
    except AttributeError:
        pass  # print(f"No Mangle R: {err}")


class LinearFunction2(MutableSequence):
    def __init__(self, init=None):
        self.__func = list()
        self.__instance = None
        if init is not None:
            for i in init:
                if isinstance(i, LinearComponent2):
                    self.__func.append(i)
                else:
                    raise TypeError(
                        f"members of {self.__class__.__name__} must be LinearComponent2"
                    )

    def set_instance(self, instance):
        self.__instance = instance

    def __get__(self, instance, owner):
        # print(f"get from descriptor object\n  instance={instance}\n  owner={owner}")
        self.__instance = instance
        return self

    def __set__(self, instance, values):
        # print(f"set in descriptor object\n  instance={instance}\n  values={values}")
        self.__init__(values)
        self.__instance = instance
        _try_mangle(self.__instance)

    def __delete__(self, instance):
        # print("deleted in descriptor object", instance)
        self.__instance = instance
        self.__func = list()
        _try_mangle(self.__instance)

    def __set_name__(self, owner, name):
        # print(f"setname in descriptor object\n  owner={owner}\n  name={name}\n  self={self}")
        self.name = name

    def __getitem__(self, item):
        return self.__func[item]

    def __setitem__(self, key, value):
        if isinstance(value, LinearComponent2):
            self.__func[key] = value
            _try_mangle(self.__instance)
        else:
            raise TypeError(
                f"members of {self.__class__.__name__} must be LinearComponent2"
            )

    def __delitem__(self, key):
        del self.__func[key]
        _try_mangle(self.__instance)

    def __len__(self):
        return len(self.__func)

    def insert(self, index, value):
        if isinstance(value, LinearComponent2):
            self.__func.insert(index, value)
            _try_mangle(self.__instance)
        else:
            raise TypeError(
                f"members of {self.__class__.__name__} must be LinearComponent2"
            )

    def __add__(self, other):
        if other == ():
            return self
        if isinstance(other, LinearFunction2):
            return type(self)(*self.__func, *other.__func)
        if isinstance(other, ParameterRef):
            other = LinearComponent2(param=other)
        if isinstance(other, LinearComponent2):
            result = type(self)(self)
            result.append(other)
            return result
        raise TypeError(f"cannot add type {type(other)} to LinearFunction")

    def __iadd__(self, other):
        if isinstance(other, ParameterRef):
            other = LinearComponent2(param=other)
        if other == ():
            return self
        elif isinstance(other, LinearFunction2):
            self.__func.extend(other)
        elif isinstance(other, LinearComponent2):
            self.append(other)
        else:
            raise TypeError(f"cannot add type {type(other)} to LinearFunction")
        return self

    def __radd__(self, other):
        return LinearFunction2([*other, *self])

    def __pos__(self):
        return self

    def __mul__(self, other):
        trial = LinearFunction2()
        for component in self:
            trial.append(component * other)
        return trial

    def __rmul__(self, other):
        trial = LinearFunction2()
        for component in self:
            trial += other * component
        return trial

    def __contains__(self, val):
        if isinstance(val, ParameterRef):
            for i in self:
                if i.param == val:
                    return True
            return False
        if isinstance(val, DataRef):
            for i in self:
                if i.data == val:
                    return True
            return False
        raise TypeError(
            "the searched for content must be of type ParameterRef or DataRef"
        )

    def _index_of(self, val):
        if isinstance(val, ParameterRef):
            for n, i in enumerate(self):
                if i.param == val:
                    return n
            raise KeyError("ParameterRef not found")
        if isinstance(val, DataRef):
            for n, i in enumerate(self):
                if i.data == val:
                    return n
            raise KeyError("DataRef not found")
        raise TypeError(
            "the searched for content must be of type ParameterRef or DataRef"
        )

    #######
    def reformat_param(self, container=None, pattern=None, repl=None, **kwargs):
        """
        Transform all the parameters in the LinearFunction.

        Parameters
        ----------
        container : str
                A format string, into which the previous parameters are formatted.
                Use this to append things to the parameter names.
        pattern : str
        repl : str
                Passed to `re.sub` with each existing parameter as the base string
                to be searched.

        Examples
        --------
        >>> from larch.roles import P,X
        >>> f = P.InVehTime * X.IVTT + P.OutOfVehTime * X.OVTT
        >>> f1 = f.reformat_param('{}_Suffix')
        >>> str(f1)
        'P.InVehTime_Suffix * X.IVTT + P.OutOfVehTime_Suffix * X.OVTT'
        >>> f2 = f.reformat_param(pattern='(Veh)', repl='Vehicle')
        >>> str(f2)
        'P.InVehicleTime * X.IVTT + P.OutOfVehicleTime * X.OVTT'

        """
        import re

        r = self.__class__()
        for i in self:
            if pattern is None:
                param = i.param
            else:
                if repl is None:
                    raise TypeError("must give repl with pattern")
                param = re.sub(pattern, repl, i.param, **kwargs)
            if container is None:
                container = "{}"
            r += LinearComponent2(
                data=i.data, param=container.format(param), scale=i.scale
            )
        try:
            r._x_ident = self._x_ident
        except AttributeError:
            pass
        return r

    def reformat_data(self, container=None, pattern=None, repl=None, **kwargs):
        """
        Transform all the data in the LinearFunction.

        Parameters
        ----------
        container : str
                A format string, into which the previous data strings are formatted.
                Use this to apply common global transforms to the data.
        pattern : str
        repl : str
                Passed to `re.sub` with each existing data string as the base string
                to be searched.
        """
        import re

        r = self.__class__()
        for i in self:
            if pattern is None:
                data = i.data
            else:
                if repl is None:
                    raise TypeError("must give repl with pattern")
                data = re.sub(pattern, repl, i.data, **kwargs)
            if container is None:
                container = "{}"
            r += LinearComponent2(
                data=container.format(data), param=i.param, scale=i.scale
            )
        try:
            r._x_ident = self._x_ident
        except AttributeError:
            pass
        return r

    def __code__(self):
        return " + ".join(f"({repr(i)})" for i in self)

    def __eq__(self, other):
        if not isinstance(other, LinearFunction2):
            return False
        if len(self) != len(other):
            return False
        for i, j in zip(self, other):
            if i != j:
                return False
        return True

    # def __getstate__(self):
    # 	state = {}
    # 	state['code'] = self.__code__()
    # 	try:
    # 		state['_x_ident'] = self._x_ident
    # 	except AttributeError:
    # 		pass
    # 	return state
    #
    # def __setstate__(self, state):
    # 	self.__init__()
    # 	if state['code']:
    # 		self.__iadd__(eval(state['code']))
    # 	if '_x_ident' in state:
    # 		self._x_ident = state['_x_ident']

    def __repr__(self):
        if len(self):
            result = " + ".join(repr(i) for i in self)
            if len(result) < 80:
                return result
            else:
                return "  " + result.replace(" + ", "\n+ ")
        return f"<Empty {self.__class__.__name__}>"

    def __xml__(
        self,
        linebreaks=False,
        lineprefix="",
        exponentiate_parameters=False,
        resolve_parameters=None,
    ):
        from xmle import Elem

        x = Elem("div", attrib={"class": "LinearFunc"})
        for n, i in enumerate(self):
            ix_ = list(
                i.__xml__(
                    exponentiate_parameter=exponentiate_parameters,
                    resolve_parameters=resolve_parameters,
                )
            )
            if linebreaks:
                if n > 0 or lineprefix:
                    ix_.insert(0, Elem("br", tail=lineprefix + " + "))
            else:
                if n < len(self) - 1:
                    if ix_[-1].tail is None:
                        ix_[-1].tail = " + "
                    else:
                        ix_[-1].tail += " + "
            for ii in ix_:
                x << ii
        return x

    def _repr_html_(self):
        return self.__xml__().tostring()

    def data(self, cls=None):
        if cls is None:
            return [_.data for _ in self]
        else:
            return [cls(_.data) for _ in self]

    def _evaluate(self, p_getter, x_namespace=None, **more_x_namespace):
        if hasattr(p_getter, "pvalue") and callable(p_getter.pvalue):
            p_getter = p_getter.pvalue
        return sum(
            j._evaluate(p_getter, x_namespace=x_namespace, **more_x_namespace)
            for j in self
        )

    def copy(self):
        result = self.__class__(self)
        return result

    def __deepcopy__(self, memodict):
        result = self.__class__()
        for i in self:
            result.append(copy.deepcopy(i, memodict))
        return result

    def _linear_plot_2d_data(
        self, p_getter, x_name, x_min, x_max, n_points=100, **other_namespace
    ):
        import numpy

        if hasattr(self, "plotting_namespace") and len(other_namespace) == 0:
            other_namespace = self.plotting_namespace
        x = numpy.linspace(x_min, x_max, n_points)
        y = self._evaluate(p_getter, {x_name: x}, **other_namespace)
        return x, y

    def linear_plot_2d(
        self,
        p_getter,
        x_name,
        x_min,
        x_max,
        n_points=100,
        *,
        xlabel=None,
        svg=True,
        header=None,
        **other_namespace,
    ):
        # Delayed evaluation mode...
        if p_getter is None:
            return lambda x: self.linear_plot_2d(
                x,
                x_name,
                x_min,
                x_max,
                n_points=n_points,
                xlabel=xlabel,
                svg=svg,
                header=header,
                **other_namespace,
            )

        # Active evaluation mode...
        from .._optional import pyplot as plt

        plt.clf()
        x, y = self._linear_plot_2d_data(
            p_getter, x_name, x_min, x_max, n_points, **other_namespace
        )
        if hasattr(self, "plotting_label"):
            plt.plot(x, y, label=self.plotting_label)
        else:
            plt.plot(x, y)
        if xlabel is None:
            plt.xlabel(x_name)
        else:
            plt.xlabel(xlabel)
        plt.tight_layout(pad=0.5)
        from .util.plotting import plot_as_svg_xhtml

        if svg is True:
            svg = {}
        if svg or svg == {}:
            if header is not None:
                svg["header"] = header
            return plot_as_svg_xhtml(plt, **svg)
        else:
            plt.show()

    def _inplot_linear_plot_2d(
        self,
        plt,
        p_getter,
        x_name,
        x_min,
        x_max,
        n_points=100,
        *,
        xlabel=None,
        svg=True,
        header=None,
        **other_namespace,
    ):
        # Delayed evaluation mode...
        if p_getter is None:
            return lambda x: self._inplot_linear_plot_2d(
                plt,
                x,
                x_name,
                x_min,
                x_max,
                n_points=n_points,
                xlabel=xlabel,
                svg=svg,
                header=header,
                **other_namespace,
            )

        # Active evaluation mode...
        x, y = self._linear_plot_2d_data(
            p_getter, x_name, x_min, x_max, n_points, **other_namespace
        )
        if hasattr(self, "plotting_label"):
            plt.plot(x, y, label=self.plotting_label)
        else:
            plt.plot(x, y)

    def total_ordering_increasing(self):
        from toolz.itertoolz import sliding_window

        snaps = []
        for windows_size in range(2, len(self) - 1):
            for sub_p in sliding_window(windows_size, self):
                snaps.append(sub_p[0].param.lessthan(sub_p[-1].param))

        return snaps


class DictOfLinearFunction2(MutableMapping):
    def __init__(self, mapping=None, alts_validator=None, **kwargs):
        self.__map = {}
        if mapping is None:
            mapping = {}
        for k, v in mapping.items():
            self.__map[k] = LinearFunction2(v)
        for k, v in kwargs.items():
            try:
                self.__map[k] = LinearFunction2(v)
            except Exception:
                print(v)
                print(type(v))
                raise

        self._alts_validator = alts_validator
        self.__instance = None

    def __get__(self, instance, owner):
        # print(f"get from descriptor object\n  instance={instance}\n  owner={owner}")
        self.__instance = instance
        return self

    def __set__(self, instance, values):
        # print(f"set in descriptor object\n  instance={instance}\n  values={values}")
        self.__init__(values)
        self.__instance = instance
        _try_mangle(self.__instance)

    def __delete__(self, instance):
        # print("deleted in descriptor object", instance)
        self.__instance = instance
        self.__func = list()
        _try_mangle(self.__instance)

    def __set_name__(self, owner, name):
        # print(f"setname in descriptor object\n  owner={owner}\n  name={name}\n  self={self}")
        self.name = name

    def set_alts_validator(self, av):
        self._alts_validator = av

    def __getitem__(self, k):
        try:
            v = self.__map[k]
            v.set_instance(self.__instance)
            return v
        except KeyError:
            if self._alts_validator is None or self._alts_validator(k):
                v = self.__map[k] = LinearFunction2()
                v.set_instance(self.__instance)
                return v
            else:
                raise

    def __setitem__(self, k, v):
        if isinstance(v, ParameterRef):
            v = v * DataRef("1")
        if isinstance(v, int) and v == 0:
            v = LinearFunction2()
        elif isinstance(v, LinearComponent2):
            v = LinearFunction2([v])
        elif isinstance(v, list):
            v = LinearFunction2(v)
        elif not isinstance(v, LinearFunction2):
            raise TypeError(f"only accepts LinearFunction2 values, not {type(v)}")
        if "with error" in repr(v):
            raise ValueError("found error here")
        v.set_instance(self.__instance)
        self.__map[k] = v
        _try_mangle(self.__instance)

    def __delitem__(self, key):
        del self.__map[key]
        _try_mangle(self.__instance)

    def __iter__(self):
        return iter(self.__map.keys())

    def __len__(self):
        return len(self.__map)

    def copy(self):
        return type(self)(self)

    def __repr__(self):
        return f"{type(self).__name__}({repr(self.__map)})"

    def __xml__(self):
        from xmle import Elem

        x = Elem("div")
        t = x.elem("table", style="margin-top:1px;")
        t.elem(
            "caption",
            text=f"<larch.{self.__class__.__name__}>",
            style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;font-style:normal;font-size:100%;padding:0px;",
        )
        if len(self):
            tr = t.elem("tr")
            tr.elem("th", text="alt")
            tr.elem("th", text="formula")
            for k, v in self.items():
                tr = t.elem("tr")
                tr.elem("td", text=str(k))
                try:
                    v_ = v.__xml__()
                except AttributeError:
                    tr.elem("td", text=str(v))
                else:
                    tr.elem("td") << v_
        else:
            tr = t.elem("tr")
            tr.elem("td", text="<empty>")
        return x

    def _repr_html_(self):
        return self.__xml__().tostring()


# ===========================================================================


class LinearComponent0(TouchNotify):
    def __init__(self, *args, param=None, data=None, scale=None, **kwargs):
        if len(args):
            if isinstance(args[0], LinearComponent0):
                super().__init__(**kwargs)
                self._param = param or args[0]._param
                self._data = data or args[0]._data
                self._scale = scale or args[0]._scale
            else:
                raise TypeError(
                    "init LinearComponent0 with LinearComponent0 or keyword arguments"
                )
        super().__init__(**kwargs)
        self._param = param or "1"
        self._data = data or "1"
        self._scale = scale or 1.0

    @property
    def data(self):
        return DataRef(self._data)

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def param(self):
        return ParameterRef(self._param)

    @param.setter
    def param(self, value):
        self._param = value

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, value):
        self._scale = value

    def __pos__(self):
        return self

    def __add__(self, other):
        if other == ():
            return self
        elif isinstance(other, (LinearFunction, LinearComponent0)):
            return (LinearFunction() + self) + other
        elif other == 0:
            return self
        else:
            raise NotImplementedError()

    def __radd__(self, other):
        if other == 0:
            return self
        return other.__add__(self)

    def __mul__(self, other):
        if isinstance(
            other,
            (
                int,
                float,
            ),
        ):
            return LinearComponent0(
                data=self.data, param=self.param, scale=self.scale * other
            )
        elif isinstance(other, (DataRef,)):
            return LinearComponent0(
                data=self.data * other, param=self.param, scale=self.scale
            )
        else:
            raise TypeError(
                f"unsupported operand type(s) for LinearComponent0: {type(other)}"
            )

    def __rmul__(self, other):
        if isinstance(
            other,
            (
                int,
                float,
            ),
        ):
            return LinearComponent0(
                data=self.data, param=self.param, scale=self.scale * other
            )
        elif isinstance(other, (DataRef,)):
            return LinearComponent0(
                data=self.data * other, param=self.param, scale=self.scale
            )
        else:
            raise TypeError(
                f"unsupported operand type(s) for LinearComponent0: {type(other)}"
            )

    def __imul__(self, other):
        if isinstance(
            other,
            (
                int,
                float,
            ),
        ):
            self.scale = self.scale * other
        elif isinstance(other, (DataRef,)):
            self.data = self.data * other
        else:
            raise TypeError(
                f"unsupported operand type(s) for LinearComponent0: {type(other)}"
            )

    def __iter__(self):
        return iter(LinearFunction() + self)

    def __eq__(self, other):
        if isinstance(other, LinearFunction) and len(other) == 1:
            other = other[0]
        if not isinstance(other, LinearComponent0):
            return False
        if self.param != other.param:
            return False
        if self.data != other.data:
            return False
        if self.scale != other.scale:
            return False
        return True

    def __repr__(self):
        try:
            if self.scale == 1.0:
                return f"({self.param!r} * {self.data!r})"
            return f"({self.param!r} * {self.scale} * {self.data!r})"
        except AttributeError:
            return f"<LinearComponent0 {id(self)} with error>"

    def __xml__(self, exponentiate_parameter=False, resolve_parameters=None):
        from xmle import Elem

        x = Elem("div")
        # x << tooltipped_style()

        if resolve_parameters is not None:
            if exponentiate_parameter:
                plabel = resolve_parameters.pvalue(
                    self.param, default_value="This is a Parameter"
                )
                if not isinstance(plabel, str):
                    plabel = f"exp({plabel:.3g})={numpy.exp(plabel):.3g}"
            else:
                plabel = "{:.3g}".format(
                    resolve_parameters.pvalue(
                        self.param, default_value="This is a Parameter"
                    )
                )
        else:
            plabel = "This is a Parameter"

        if self.scale == 1.0:
            if exponentiate_parameter:
                x.elem("span", tail="exp(")
                a_p = x.elem(
                    "div",
                    attrib={"class": "tooltipped"},
                    text=repr(self.param),
                    tail=") * ",
                )
                a_p.elem("span", attrib={"class": "tooltiptext"}, text=plabel)
            else:
                a_p = x.elem(
                    "div",
                    attrib={"class": "tooltipped"},
                    text=repr(self.param),
                    tail=" * ",
                )
                a_p.elem("span", attrib={"class": "tooltiptext"}, text=plabel)
        else:
            if exponentiate_parameter:
                x.elem("span", tail="exp(")
                a_p = x.elem(
                    "div",
                    attrib={"class": "tooltipped"},
                    text=repr(self.param),
                    tail=f" * {self.scale}) * ",
                )
                a_p.elem("span", attrib={"class": "tooltiptext"}, text=plabel)
            else:
                a_p = x.elem(
                    "div",
                    attrib={"class": "tooltipped"},
                    text=repr(self.param),
                    tail=f" * {self.scale} * ",
                )
                a_p.elem("span", attrib={"class": "tooltiptext"}, text=plabel)
        a_x = x.elem("div", attrib={"class": "tooltipped"}, text=repr(self.data))
        a_x.elem("span", attrib={"class": "tooltiptext"}, text="This is Data")
        return x

    def _repr_html_(self):
        return self.__xml__().tostring()

    def _evaluate(self, p_getter, x_namespace=None, **kwargs):
        return (
            self._scale
            * p_getter(str(self.param))
            * self.data.eval(namespace=x_namespace, **kwargs)
        )

    def __copy__(self):
        return LinearComponent0(
            param=self._param,
            data=self._data,
            scale=self._scale,
        )

    def __deepcopy__(self, memodict):
        return LinearComponent0(
            param=copy.deepcopy(
                self._param if isinstance(self._param, str) else str(self._param),
                memodict,
            ),
            data=copy.deepcopy(
                self._data if isinstance(self._data, str) else str(self._data), memodict
            ),
            scale=copy.deepcopy(self._scale, memodict),
        )


# ===========================================================================


class LinearFunction0(TouchNotify, list):
    def __init__(self, *args, **kwargs):
        if len(args) == 1 and args[0] is None:
            args = ()
        super().__init__(*args, **kwargs)

    def __add__(self, other):
        if other == ():
            return self
        if isinstance(other, LinearFunction0):
            return type(self)(super().__add__(other))
        if isinstance(other, ParameterRef):
            other = LinearComponent0(param=other)
        if isinstance(other, LinearComponent0):
            result = type(self)(self)
            result.append(other)
            return result
        raise TypeError(f"cannot add type {type(other)} to LinearFunction0")

    def __iadd__(self, other):
        if isinstance(other, ParameterRef):
            other = LinearComponent0(param=other)
        if other == ():
            return self
        elif isinstance(other, LinearFunction0):
            super().__iadd__(other)
        elif isinstance(other, LinearComponent0):
            super().append(other)
        else:
            raise TypeError(f"cannot add type {type(other)} to LinearFunction0")
        return self

    def __radd__(self, other):
        return LinearFunction0() + other + self

    def __pos__(self):
        return self

    def __mul__(self, other):
        trial = LinearFunction0()
        for component in self:
            trial += component * other
        return trial

    def __rmul__(self, other):
        trial = LinearFunction0()
        for component in self:
            trial += other * component
        return trial

    # def evaluate(self, dataspace, model):
    # 	if len(self) > 0:
    # 		i = self[0]
    # 		y = i.data.eval(**dataspace) * i.param.default_value(0).value(model)
    # 	for i in self[1:]:
    # 		y += i.data.eval(**dataspace) * i.param.default_value(0).value(model)
    # 	return y
    #
    # def evaluator1d(self, factorlabel='', x_ident=None):
    # 	if x_ident is None:
    # 		try:
    # 			x_ident = self._x_ident
    # 		except AttributeError:
    # 			raise TypeError('a x_ident must be given')
    # 	if x_ident is None:
    # 		raise TypeError('a x_ident must be given')
    # 	from .util.plotting import ComputedFactor
    # 	return ComputedFactor(label=factorlabel, func=lambda x, m: self.evaluate({x_ident: x}, m))

    def __contains__(self, val):
        if isinstance(val, ParameterRef):
            for i in self:
                if i.param == val:
                    return True
            return False
        if isinstance(val, DataRef):
            for i in self:
                if i.data == val:
                    return True
            return False
        raise TypeError(
            "the searched for content must be of type ParameterRef or DataRef"
        )

    def _index_of(self, val):
        if isinstance(val, ParameterRef):
            for n, i in enumerate(self):
                if i.param == val:
                    return n
            raise KeyError("ParameterRef not found")
        if isinstance(val, DataRef):
            for n, i in enumerate(self):
                if i.data == val:
                    return n
            raise KeyError("DataRef not found")
        raise TypeError(
            "the searched for content must be of type ParameterRef or DataRef"
        )

    def reformat_param(self, container=None, pattern=None, repl=None, **kwargs):
        """
        Transform all the parameters in the LinearFunction0.

        Parameters
        ----------
        container : str
                A format string, into which the previous parameters are formatted.
                Use this to append things to the parameter names.
        pattern : str
        repl : str
                Passed to `re.sub` with each existing parameter as the base string
                to be searched.

        Examples
        --------
        >>> from larch.roles import P,X
        >>> f = P.InVehTime * X.IVTT + P.OutOfVehTime * X.OVTT
        >>> f1 = f.reformat_param('{}_Suffix')
        >>> str(f1)
        'P.InVehTime_Suffix * X.IVTT + P.OutOfVehTime_Suffix * X.OVTT'
        >>> f2 = f.reformat_param(pattern='(Veh)', repl='Vehicle')
        >>> str(f2)
        'P.InVehicleTime * X.IVTT + P.OutOfVehicleTime * X.OVTT'

        """
        import re

        r = LinearFunction0()
        for i in self:
            if pattern is None:
                param = i.param
            else:
                if repl is None:
                    raise TypeError("must give repl with pattern")
                param = re.sub(pattern, repl, i.param, **kwargs)
            if container is None:
                container = "{}"
            r += LinearComponent0(data=i.data, param=container.format(param))
        try:
            r._x_ident = self._x_ident
        except AttributeError:
            pass
        return r

    def reformat_data(self, container=None, pattern=None, repl=None, **kwargs):
        """
        Transform all the data in the LinearFunction0.

        Parameters
        ----------
        container : str
                A format string, into which the previous data strings are formatted.
                Use this to apply common global transforms to the data.
        pattern : str
        repl : str
                Passed to `re.sub` with each existing data string as the base string
                to be searched.
        """
        import re

        r = LinearFunction0()
        for i in self:
            if pattern is None:
                data = i.data
            else:
                if repl is None:
                    raise TypeError("must give repl with pattern")
                data = re.sub(pattern, repl, i.data, **kwargs)
            if container is None:
                container = "{}"
            r += LinearComponent0(data=container.format(data), param=i.param)
        try:
            r._x_ident = self._x_ident
        except AttributeError:
            pass
        return r

    def __code__(self):
        return " + ".join(f"({repr(i)})" for i in self)

    def __eq__(self, other):
        if not isinstance(other, LinearFunction0):
            return False
        if len(self) != len(other):
            return False
        for i, j in zip(self, other):
            if i != j:
                return False
        return True

    def __getstate__(self):
        state = {}
        state["code"] = self.__code__()
        try:
            state["_x_ident"] = self._x_ident
        except AttributeError:
            pass
        return state

    def __setstate__(self, state):
        self.__init__()
        if state["code"]:
            self.__iadd__(eval(state["code"]))
        if "_x_ident" in state:
            self._x_ident = state["_x_ident"]

    def __repr__(self):
        result = " + ".join(repr(i) for i in self)
        if len(result) < 80:
            return result
        else:
            return "  " + result.replace(" + ", "\n+ ")

    def __xml__(
        self,
        linebreaks=False,
        lineprefix="",
        exponentiate_parameters=False,
        resolve_parameters=None,
    ):
        from xmle import Elem

        x = Elem("div", attrib={"class": "LinearFunc"})
        for n, i in enumerate(self):
            ix_ = list(
                i.__xml__(
                    exponentiate_parameter=exponentiate_parameters,
                    resolve_parameters=resolve_parameters,
                )
            )
            if linebreaks:
                if n > 0 or lineprefix:
                    ix_.insert(0, Elem("br", tail=lineprefix + " + "))
            else:
                if n < len(self) - 1:
                    if ix_[-1].tail is None:
                        ix_[-1].tail = " + "
                    else:
                        ix_[-1].tail += " + "
            for ii in ix_:
                x << ii
        return x

    def _repr_html_(self):
        return self.__xml__().tostring()

    def data(self, cls=None):
        if cls is None:
            return [_.data for _ in self]
        else:
            return [cls(_.data) for _ in self]

    def set_touch_callback(self, callback):
        super().set_touch_callback(callback)
        for i in self:
            i.set_touch_callback(callback)

    def _evaluate(self, p_getter, x_namespace=None, **more_x_namespace):
        if hasattr(p_getter, "pvalue") and callable(p_getter.pvalue):
            p_getter = p_getter.pvalue
        return sum(
            j._evaluate(p_getter, x_namespace=x_namespace, **more_x_namespace)
            for j in self
        )

    def copy(self):
        result = LinearFunction0() + self
        return result

    def __deepcopy__(self, memodict):
        result = LinearFunction0()
        for i in self:
            result.append(copy.deepcopy(i, memodict))
        return result

    def _linear_plot_2d_data(
        self, p_getter, x_name, x_min, x_max, n_points=100, **other_namespace
    ):
        import numpy

        if hasattr(self, "plotting_namespace") and len(other_namespace) == 0:
            other_namespace = self.plotting_namespace
        x = numpy.linspace(x_min, x_max, n_points)
        y = self._evaluate(p_getter, {x_name: x}, **other_namespace)
        return x, y

    def linear_plot_2d(
        self,
        p_getter,
        x_name,
        x_min,
        x_max,
        n_points=100,
        *,
        xlabel=None,
        svg=True,
        header=None,
        **other_namespace,
    ):
        # Delayed evaluation mode...
        if p_getter is None:
            return lambda x: self.linear_plot_2d(
                x,
                x_name,
                x_min,
                x_max,
                n_points=n_points,
                xlabel=xlabel,
                svg=svg,
                header=header,
                **other_namespace,
            )

        # Active evaluation mode...
        from .._optional import pyplot as plt

        plt.clf()
        x, y = self._linear_plot_2d_data(
            p_getter, x_name, x_min, x_max, n_points, **other_namespace
        )
        if hasattr(self, "plotting_label"):
            plt.plot(x, y, label=self.plotting_label)
        else:
            plt.plot(x, y)
        if xlabel is None:
            plt.xlabel(x_name)
        else:
            plt.xlabel(xlabel)
        plt.tight_layout(pad=0.5)
        from .util.plotting import plot_as_svg_xhtml

        if svg is True:
            svg = {}
        if svg or svg == {}:
            if header is not None:
                svg["header"] = header
            return plot_as_svg_xhtml(plt, **svg)
        else:
            plt.show()

    def _inplot_linear_plot_2d(
        self,
        plt,
        p_getter,
        x_name,
        x_min,
        x_max,
        n_points=100,
        *,
        xlabel=None,
        svg=True,
        header=None,
        **other_namespace,
    ):
        # Delayed evaluation mode...
        if p_getter is None:
            return lambda x: self._inplot_linear_plot_2d(
                plt,
                x,
                x_name,
                x_min,
                x_max,
                n_points=n_points,
                xlabel=xlabel,
                svg=svg,
                header=header,
                **other_namespace,
            )

        # Active evaluation mode...
        x, y = self._linear_plot_2d_data(
            p_getter, x_name, x_min, x_max, n_points, **other_namespace
        )
        if hasattr(self, "plotting_label"):
            plt.plot(x, y, label=self.plotting_label)
        else:
            plt.plot(x, y)

    def total_ordering_increasing(self):
        from toolz.itertoolz import sliding_window

        snaps = []
        for windows_size in range(2, len(self) - 1):
            for sub_p in sliding_window(windows_size, self):
                snaps.append(sub_p[0].param.lessthan(sub_p[-1].param))

        return snaps


# ====================================================================================================
def multiple_linear_plot_2d(
    linear_funcs,
    p_getter,
    x_name,
    x_min,
    x_max,
    n_points=100,
    *,
    xlabel=None,
    svg=True,
    header=None,
    other_namespace=None,
):
    # Delayed evaluation mode...
    if p_getter is None:
        return lambda x: multiple_linear_plot_2d(
            linear_funcs,
            x,
            x_name,
            x_min,
            x_max,
            n_points=n_points,
            xlabel=xlabel,
            svg=svg,
            header=header,
            other_namespace=other_namespace,
        )

    # Active evaluation mode...
    from .._optional import pyplot as plt

    plt.clf()

    if other_namespace is None:
        for lf in linear_funcs:
            lf._inplot_linear_plot_2d(
                plt,
                p_getter,
                x_name,
                x_min,
                x_max,
                n_points,
                xlabel=xlabel,
                svg=svg,
                header=header,
            )
    else:
        for lf, oname in zip(linear_funcs, other_namespace):
            lf._inplot_linear_plot_2d(
                plt,
                p_getter,
                x_name,
                x_min,
                x_max,
                n_points,
                xlabel=xlabel,
                svg=svg,
                header=header,
                **oname,
            )

    if xlabel is not None:
        plt.xlabel(xlabel)

    plt.legend()

    plt.tight_layout(pad=0.5)
    from .util.plotting import plot_as_svg_xhtml

    if svg is True:
        svg = {}
    if svg or svg == {}:
        if header is not None:
            svg["header"] = header
        return plot_as_svg_xhtml(plt, **svg)
    else:
        plt.show()


# ====================================================================================================


_RaiseKeyError = object()  # singleton for no-default behavior


class DictOfLinearFunction0(TouchNotify, dict):
    @staticmethod  # because this doesn't make sense as a global function.
    def _process_args(mapping=(), **kwargs):
        if hasattr(mapping, "items"):
            mapping = mapping.items()
        return ((k, v) for k, v in chain(mapping, kwargs.items()))

    def __init__(self, mapping=(), alts_validator=None, **kwargs):
        if mapping is None:
            mapping = ()
        super().__init__(self._process_args(mapping, **kwargs))
        self._alts_validator = alts_validator

    def set_alts_validator(self, av):
        self._alts_validator = av

    def __getitem__(self, k):
        try:
            return super().__getitem__(k)
        except KeyError:
            if self._alts_validator is None or self._alts_validator(k):
                super().__setitem__(k, LinearFunction0())
                return super().__getitem__(k)
            else:
                raise

    def __setitem__(self, k, v):
        if isinstance(v, ParameterRef):
            v = v * DataRef("1")
        if isinstance(v, int) and v == 0:
            v = LinearFunction0()
        elif isinstance(v, LinearComponent0):
            v = LinearFunction0() + v
        elif isinstance(v, LinearFunction0):
            pass
        elif isinstance(v, list):
            v_ = LinearFunction0()
            for vv in v:
                v_ = v_ + vv
            v = v_
        else:
            raise TypeError("only accepts LinearFunction0 values")
        if "with error" in repr(v):
            raise ValueError("found error here")
        return super().__setitem__(k, v)

    def setdefault(self, k, default=None):
        if default is None:
            default = LinearFunction0()
        if isinstance(default, LinearComponent0):
            default = LinearFunction0() + default
        elif isinstance(default, LinearFunction0):
            pass
        else:
            raise TypeError("only accepts LinearFunction0 values")
        return super().setdefault(k, default)

    def copy(self):  # don't delegate w/ super - dict.copy() -> dict :(
        return type(self)(self)

    def copy_without_touch_callback(self, deep=True):
        result = type(self)()
        if deep:
            for k in self.keys():
                result[k] = list(self[k])
        else:
            for k in self.keys():
                result[k] = self[k]
        return result

    def __repr__(self):
        return f"{type(self).__name__}({super().__repr__()})"

    def set_touch_callback(self, callback):
        super().set_touch_callback(callback)
        for i in self.values():
            i.set_touch_callback(callback)

    def __xml__(self):
        from xmle import Elem

        x = Elem("div")
        t = x.elem("table", style="margin-top:1px;")
        t.elem(
            "caption",
            text=f"<larch.{self.__class__.__name__}>",
            style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;font-style:normal;font-size:100%;padding:0px;",
        )
        if len(self):
            tr = t.elem("tr")
            tr.elem("th", text="alt")
            tr.elem("th", text="formula")
            for k, v in self.items():
                tr = t.elem("tr")
                tr.elem("td", text=str(k))
                try:
                    v_ = v.__xml__()
                except AttributeError:
                    tr.elem("td", text=str(v))
                else:
                    tr.elem("td") << v_
        else:
            tr = t.elem("tr")
            tr.elem("td", text="<empty>")
        return x

    def _repr_html_(self):
        return self.__xml__().tostring()


#
#
#


class DictOfStrings(TouchNotify, dict):
    @staticmethod  # because this doesn't make sense as a global function.
    def _process_args(mapping=(), **kwargs):
        if hasattr(mapping, "items"):
            mapping = mapping.items()
        return ((k, v) for k, v in chain(mapping, kwargs.items()))

    def __init__(self, mapping=(), **kwargs):
        if mapping is None:
            mapping = ()
        super().__init__(self._process_args(mapping, **kwargs))

    def __contains__(self, k):
        return super().__contains__(str(k))

    def __getitem__(self, k):
        return super().__getitem__(str(k))

    def __setitem__(self, k, v):
        return super().__setitem__(str(k), str(v))

    def setdefault(self, k, default=""):
        return super().setdefault(str(k), str(default))

    def copy(self):  # don't delegate w/ super - dict.copy() -> dict :(
        return type(self)(self)

    def __repr__(self):
        return f"{type(self).__name__}({super().__repr__()})"

    def __xml__(self):
        from xmle import Elem

        x = Elem("div")
        t = x.elem("table", style="margin-top:1px;")
        t.elem(
            "caption",
            text=f"<larch.{self.__class__.__name__}>",
            style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;font-style:normal;font-size:100%;padding:0px;",
        )
        if len(self):
            tr = t.elem("tr")
            tr.elem("th", text="key")
            tr.elem("th", text="value")
            for k, v in self.items():
                tr = t.elem("tr")
                tr.elem("td", text=str(k))
                tr.elem("td", text=str(v))
        else:
            tr = t.elem("tr")
            tr.elem("td", text="<empty>")
        return x

    def _repr_html_(self):
        return self.__xml__().tostring()


class DictOfStringKeys(TouchNotify, dict):
    @staticmethod  # because this doesn't make sense as a global function.
    def _process_args(mapping=(), **kwargs):
        if hasattr(mapping, "items"):
            mapping = mapping.items()
        return ((k, v) for k, v in chain(mapping, kwargs.items()))

    def __init__(self, mapping=(), **kwargs):
        if mapping is None:
            mapping = ()
        super().__init__(self._process_args(mapping, **kwargs))

    def __contains__(self, k):
        return super().__contains__(str(k))

    def __getitem__(self, k):
        return super().__getitem__(str(k))

    def __setitem__(self, k, v):
        return super().__setitem__(str(k), v)

    def setdefault(self, k, default=""):
        return super().setdefault(str(k), default)

    def copy(self):  # don't delegate w/ super - dict.copy() -> dict :(
        return type(self)(self)

    def __repr__(self):
        return f"{type(self).__name__}({super().__repr__()})"

    def __xml__(self):
        from xmle import Elem

        x = Elem("div")
        t = x.elem("table", style="margin-top:1px;")
        t.elem(
            "caption",
            text=f"<larch.{self.__class__.__name__}>",
            style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;font-style:normal;font-size:100%;padding:0px;",
        )
        if len(self):
            tr = t.elem("tr")
            tr.elem("th", text="key")
            tr.elem("th", text="value")
            for k, v in self.items():
                tr = t.elem("tr")
                tr.elem("td", text=str(k))
                tr.elem("td", text=str(v))
        else:
            tr = t.elem("tr")
            tr.elem("td", text="<empty>")
        return x

    def _repr_html_(self):
        return self.__xml__().tostring()



import logging
from collections import OrderedDict

import numpy
import pandas

from ..warning import ignore_warnings

logger = logging.getLogger(__name__)

try:
    from ipywidgets import FloatProgress as ipywidgets_FloatProgress
except ImportError:
    ipywidgets_FloatProgress = ()


def debug_array(name, arr):
    print(f"<{name} shape={arr.shape}>")
    print(arr)
    print(f"</{name}>")


def _approx_fprime_helper(xk, f, epsilon, args=(), f0=None, *, status_widget=None):
    """See ``approx_fprime``.  An optional initial function value arg is added."""
    if f0 is None:
        f0_ = f(*((xk,) + args))
        try:
            f0 = f0_.copy()
        except AttributeError:
            f0 = f0_
    f0 = numpy.nan_to_num(f0)
    try:
        f_shape = f0.shape
    except AttributeError:
        f_shape = ()
    grad = numpy.zeros((len(xk),) + f_shape, float)
    ei = numpy.zeros((len(xk),), float)
    for k in range(len(xk)):
        ei[k] = 1.0
        d = epsilon * ei
        grad[k] = (numpy.nan_to_num(f(*((xk + d,) + args))) - f0) / d[k]
        ei[k] = 0.0
        if status_widget:
            status_widget(f"{k} / {len(xk)}")
    return grad


def _approx_fprime_helper_trailing(xk, f, epsilon, args=(), f0=None):
    """See ``approx_fprime``.  An optional initial function value arg is added."""
    if f0 is None:
        f0_ = f(*((xk,) + args))
        try:
            f0 = f0_.copy()
        except AttributeError:
            f0 = f0_
    numpy.nan_to_num(f0, copy=False)
    try:
        f_shape = f0.shape
    except AttributeError:
        f_shape = ()
    grad = numpy.zeros(f_shape + (len(xk),), float)
    ei = numpy.zeros((len(xk),), float)
    for k in range(len(xk)):
        ei[k] = 1.0
        d = epsilon * ei
        grad[..., k] = (numpy.nan_to_num(f(*((xk + d,) + args))) - f0) / d[k]
        ei[k] = 0.0
    return grad


def approx_fprime(xk, f, epsilon=None, trailing=False, *args, status_widget=None):
    """Finite-difference approximation of the gradient of a scalar function.

    Parameters
    ----------
    xk : array_like
        The coordinate vector at which to determine the gradient of `f`.
    f : callable
        The function of which to determine the gradient (partial derivatives).
        Should take `xk` as first argument, other arguments to `f` can be
        supplied in ``*args``.  Should return a scalar, the value of the
        function at `xk`.
    epsilon : array_like
        Increment to `xk` to use for determining the function gradient.
        If a scalar, uses the same finite difference delta for all partial
        derivatives.  If an array, should contain one value per element of
        `xk`.
    *args : args, optional
        Any other arguments that are to be passed to `f`.

    Returns
    -------
    grad : ndarray
        The partial derivatives of `f` to `xk`.

    See Also
    --------
    check_grad : Check correctness of gradient function against approx_fprime.

    Notes
    -----
    The function gradient is determined by the forward finite difference
    formula::

                 f(xk[i] + epsilon[i]) - f(xk[i])
        f'[i] = ---------------------------------
                            epsilon[i]

    The main use of `approx_fprime` is in scalar function optimizers like
    `fmin_bfgs`, to determine numerically the Jacobian of a function.

    """
    if epsilon is None:
        epsilon = numpy.sqrt(numpy.finfo(float).eps)
    if trailing:
        return _approx_fprime_helper_trailing(xk, f, epsilon, args=args)
    else:
        return _approx_fprime_helper(
            xk, f, epsilon, args=args, status_widget=status_widget
        )


def similarity(a, b, to_zero=None):
    """
    Compute the similarity between two values or arrays.

    Returns a similarity measure indicating approximately the number of
    significant figures in common between the two values.  Returns 100
    if the values are exactly equal.

    Parameters
    ----------
    a,b : numeric or arrays of same size
        Scalar values are converted to numpy arrays.
    to_zero : numeric, optional
        A threshold level for values that can be
        considered to match zero, where if one value is
        exactly zero and the other is not, the similarity
        is equal to the inverse absolute value of the other value
        multiplied by `to_zero`

    Returns
    -------
    ndarray

    """
    a = numpy.asarray(a)
    b = numpy.asarray(b)
    magnitude = numpy.fmax(numpy.fabs(a), numpy.fabs(b))
    difference = numpy.abs(a - b)
    with numpy.errstate(divide="ignore", invalid="ignore"):
        similar = -numpy.log10(difference / magnitude)
    similar[a == b] = 100
    if to_zero:
        with ignore_warnings():
            similar[a == 0] = (to_zero / difference)[a == 0]
            similar[b == 0] = (to_zero / difference)[b == 0]
    return similar


def _color_poor_similarity(val):
    """
    Take a scalar similarity and returns a string.

    Results in a string with
    the css property `'color: red'` for negative
    strings, black otherwise.
    """
    if val < 3:
        return "color: #FF0000"
    if val < 4.5:
        return "color: #BB0000"
    if val < 6:
        return "color: #880000"
    return "color: black"


def check_gradient(
    func,
    grad,
    x0,
    epsilon=None,
    names=None,
    require_sim=None,
    stylize=True,
    skip_zeros=True,
    trailing=False,
):
    """Check the gradient.

    Check the correctness of a gradient function by comparing it against a
    (forward) finite-difference approximation of the gradient.

    Parameters
    ----------
    func : callable ``func(x0, *args)``
        Function whose derivative is to be checked.
    grad : callable ``grad(x0, *args)``
        Gradient of `func`.
    x0 : ndarray
        Points to check `grad` against forward difference approximation of grad
        using `func`.
    args : *args, optional
        Extra arguments passed to `func` and `grad`.
    epsilon : float, optional
        Step size used for the finite difference approximation. It defaults to
        ``sqrt(numpy.finfo(float).eps)``, which is approximately 1.49e-08.
    names : list or pandas.Index
        A list of names by position in the returned gradient.
    require_sim : float, optional
        If any item has similarity less than this, a GradientCheckError is raise

    Returns
    -------
    err : float
        The square root of the sum of squares (i.e. the 2-norm) of the
        difference between ``grad(x0, *args)`` and the finite difference
        approximation of `grad` using func at the points `x0`.
    """
    if epsilon is None:
        epsilon = numpy.sqrt(numpy.finfo(float).eps)
    g_f = approx_fprime(x0, func, epsilon, trailing=trailing)
    g_a = grad(x0)

    g_f_ = g_f.reshape(-1)
    if isinstance(g_a, pandas.Series):
        g_a = g_a.values
    g_a_ = g_a.reshape(-1)
    lister = [
        # ('x', x0),
        ("finite_diff", g_f_),
        ("analytic", g_a_),
        ("diff", g_f_ - g_a_),
        ("similarity", similarity(g_f_, g_a_, to_zero=0.01)),
    ]
    try:
        f = pandas.DataFrame.from_dict(OrderedDict(lister))
    except Exception:
        print(lister)
        raise
    if names is not None and len(names) == g_a_.size:
        f.index = names
    elif names is not None and len(names) ** 2 == g_a_.size:
        from itertools import product

        f.index = [f"{n1}, {n2}" for n1, n2 in product(names, repeat=2)]
    elif names is not None:
        try:
            f.index = numpy.broadcast_to(names, g_f.shape).reshape(-1)
        except ValueError:
            f.index = numpy.broadcast_to(names, g_f.T.shape).T.reshape(-1)

    if skip_zeros:
        f = f[(f["finite_diff"] != 0) | (f["analytic"] != 0)]

    if not stylize:
        return f
    try:
        return f.style.map(_color_poor_similarity, subset="similarity")
    except ValueError:
        return f


def assert_gradient_check(*args, min_similarity=4, **kwargs):
    kwargs["stylize"] = False
    df = check_gradient(*args, **kwargs)
    assert df["similarity"].min() >= min_similarity



import os
import subprocess


def make_estimation_data(directory="test"):
    if os.path.exists(directory):
        check_file = os.path.join(directory, ".gitignore")
        if not os.path.exists(check_file):
            import importlib

            asim_create = importlib.import_module("activitysim.cli.create")
            asim_create.get_example("example_estimation_sf", directory)
            os.chdir(directory)
            _cp = subprocess.run(
                [
                    "activitysim",
                    "run",
                    "-c",
                    "configs_estimation/configs",
                    "-c",
                    "configs",
                    "-o",
                    "output",
                    "-d",
                    "data_sf",
                ],
                capture_output=True,
            )
            with open(check_file) as f:
                f.write("**/*.csv\n")
                f.write("**/*.txt\n")
                f.write("**/*.yaml\n")
                f.write("**/.gitignore\n")
        else:
            print(f"using existing directory `{directory}`")
            os.chdir(directory)



import logging
import warnings

import numpy as np
import pandas as pd
from xarray import Dataset

from .._optional import jax, jnp
from ..compiled import compiledmethod, jitmethod, reset_compiled_methods
from ..folding import fold_dataset
from ..optimize import OptimizeMixin
from .numbamodel import NumbaModel

logger = logging.getLogger(__name__)


def _get_jnp_array(dataset, name):
    if name not in dataset:
        return None
    return jnp.asarray(dataset[name])


def _as_jnp_array(obj):
    if obj is None:
        return None
    return jnp.asarray(obj)


class PanelMixin:
    def __init__(self, *args, **kwargs):
        self._groupid = kwargs.pop("groupid", None)

    @property
    def groupid(self):
        return self._groupid

    @groupid.setter
    def groupid(self, g):
        if g is None or isinstance(g, str):
            if self._groupid != g:
                self.mangle()
        else:
            self.mangle()
        self._groupid = g


class MangleOnChange:
    def __init__(self, *req_type):
        self.req_type = req_type

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_" + name

    def __get__(self, instance, owner):
        return getattr(instance, self.private_name)

    def __set__(self, instance, value):
        if self.req_type == (bool,):
            value = bool(value)
        elif self.req_type:
            if not isinstance(value, self.req_type):
                raise TypeError(
                    f"attribute `{self.name}` must be of type {self.req_type} not {type(value)}"
                )
        old_value = getattr(instance, self.private_name, "<--missing-->")
        if old_value == value:
            return
        else:
            setattr(instance, self.private_name, value)
            try:
                instance.mangle()
            except AttributeError:
                pass


class Model(NumbaModel, OptimizeMixin, PanelMixin):
    def __init__(self, *args, **kwargs):
        PanelMixin.__init__(self, *args, **kwargs)
        super().__init__(*args, **kwargs)
        self._n_draws = 100
        self._draws = None
        self.prerolled_draws = True
        self.common_draws = False

    @property
    def compute_engine(self):
        engine = self._compute_engine
        if engine is None:
            if not jax:
                engine = "numba"
            else:
                engine = "jax"
        return engine

    @compute_engine.setter
    def compute_engine(self, engine):
        if engine not in {"numba", "jax", None}:
            raise ValueError("invalid compute engine")
        self._compute_engine = engine
        if self._compute_engine == "jax" and not jax:
            warnings.warn(
                "jax is not installed, falling back to numba",
                stacklevel=2,
            )
            self._compute_engine = "numba"
        if self._compute_engine == "jax" and self.use_streaming:
            warnings.warn(
                "setting use_streaming to False, jax is not yet compatible",
                stacklevel=2,
            )
            self.use_streaming = False

    prerolled_draws = MangleOnChange(bool)
    common_draws = MangleOnChange(bool)
    n_draws = MangleOnChange(int)
    seed = MangleOnChange()

    # @property
    # def n_draws(self):
    #     return self._n_draws
    #
    # @n_draws.setter
    # def n_draws(self, n):
    #     if n == self._n_draws:
    #         return
    #     else:
    #         self._n_draws = n
    #         self.mangle()

    def mangle(self, data=True, structure=True):
        super().mangle(data, structure)
        self._draws = None
        reset_compiled_methods(self)

    def unmangle(self, force=False, structure_only=False):
        if not self._mangled and not force:
            return
        marker = f"_currently_unmangling_{__file__}"
        if getattr(self, marker, False):
            return
        try:
            setattr(self, marker, True)
            super().unmangle(force=force, structure_only=structure_only)
            for mix in self.mixtures:
                mix.prep(self._parameter_bucket)
            if not structure_only:
                if self.groupid is not None and self.dataset is not None:
                    self.dataset = fold_dataset(self.dataset, self.groupid)
        finally:
            delattr(self, marker)

    def reflow_data_arrays(self):
        """Reload the internal data_arrays so they are consistent with the datatree."""
        if self.compute_engine != "jax":
            return super().reflow_data_arrays()

        if self.graph is None:
            self._data_arrays = None
            return

        if self._should_preload_data:
            datatree = self.datatree
        else:
            datatree = self.datatree.replace_datasets(
                {
                    self.datatree.root_node_name: self.datatree.root_dataset.isel(
                        {self.datatree.CASEID: slice(0, 1)}
                    )
                }
            )
        if datatree is not None:
            request = self.required_data()
            if isinstance(self.groupid, str):
                request["group_co"] = self.groupid

            from .data_arrays import prepare_data

            dataset, self.dataflows = prepare_data(
                datasource=datatree,
                request=request,
                float_dtype=self.float_dtype,
                cache_dir=datatree.cache_dir,
                flows=getattr(self, "dataflows", None),
                make_unused_flows=self.use_streaming,
            )
            if isinstance(self.groupid, str):
                dataset = fold_dataset(dataset, "group")
            elif self.groupid is not None:
                dataset = fold_dataset(dataset, self.groupid)
            self.dataset = dataset
            try:
                self._data_arrays = self.dataset.dc.to_arrays(
                    self.graph,
                    float_dtype=self.float_dtype,
                )
            except KeyError:  # no defined caseid dimension, JAX only
                self._data_arrays = None
                self.work_arrays = None
            else:
                if self.work_arrays is not None:
                    self._rebuild_work_arrays()

    @property
    def data_as_loaded(self):
        return self._dataset

    @property
    def dataset(self):
        """larch.Dataset : Data arrays as loaded for model computation."""
        super().unmangle()
        if self._dataset is None:
            self.reflow_data_arrays()
        try:
            return self._dataset
        except AttributeError:
            return None

    @dataset.setter
    def dataset(self, dataset):
        if dataset is self._dataset:
            return
        # from xarray import Dataset as _Dataset

        if isinstance(dataset, Dataset):
            if self.groupid is not None:
                dataset = fold_dataset(dataset, self.groupid)
            self._dataset = dataset
            self._data_arrays = None
            self._rebuild_fixed_arrays()
        # elif isinstance(dataset, _Dataset):
        #     if self.groupid is not None:
        #         dataset = fold_dataset(dataset, self.groupid)
        #     self._dataset = Dataset(dataset)
        #     self._data_arrays = None
        #     self._rebuild_fixed_arrays()
        else:
            raise TypeError(f"dataset must be Dataset not {type(dataset)}")

    def make_random_draws(self, engine="numpy"):
        self.unmangle()
        for i in self.mixtures:
            i.prep(self._parameter_bucket)
        n_panels = self.dataset.dc.n_panels
        n_mixtures = len(self.mixtures)
        n_draws = self.n_draws
        draws = None
        if self._draws is not None:
            if self.common_draws and self._draws.shape == (n_draws, n_mixtures):
                draws = self._draws
            if not self.common_draws and self._draws.shape == (
                n_panels,
                n_draws,
                n_mixtures,
            ):
                draws = self._draws
        if draws is None:
            if engine == "numpy":
                seed = getattr(self, "seed", 0)
                if self.common_draws:
                    if n_draws > 0 and n_mixtures > 0:
                        draws, seed = self._make_random_draws_numpy(
                            n_draws, n_mixtures, seed
                        )
                else:
                    if n_draws > 0 and n_mixtures > 0 and n_panels > 0:
                        draws, seed = self._make_random_draws_numpy_2(
                            n_draws, n_mixtures, n_panels, seed
                        )
                    else:
                        draws = None
            elif engine == "jax":
                seed = getattr(self, "seed", 0)
                rk = jax.random.PRNGKey(seed)
                if self.common_draws:
                    if n_draws > 0 and n_mixtures > 0:
                        draws = self._make_random_draws_out(n_draws, n_mixtures, rk)[0]
                else:
                    if n_draws > 0 and n_mixtures > 0 and n_panels > 0:
                        draws = self._make_random_draws_out_2(
                            n_draws, n_mixtures, n_panels, rk
                        )[0]
                    else:
                        draws = None
            else:
                raise ValueError(f"unknown random engine {engine!r}")
            if self.prerolled_draws:
                self._draws = draws
        return draws

    def _make_random_draws_numpy(self, n_draws, n_mixtures, seed):
        if isinstance(seed, np.random.Generator):
            rgen = seed
        else:
            rgen = np.random.default_rng(seed)
        draws = (
            rgen.random(size=[n_draws, n_mixtures], dtype=np.float32)
            + np.arange(n_draws, dtype=np.float32)[:, np.newaxis]
        )
        for i in range(n_mixtures):
            rgen.shuffle(draws[:, i])
        return np.clip(draws / n_draws, 0, np.float32(1 - 1e-7)), rgen

    def _make_random_draws_numpy_2(self, n_draws, n_mixtures, n_panels, seed):
        if isinstance(seed, np.random.Generator):
            rgen = seed
        else:
            rgen = np.random.default_rng(seed)
        draws = (
            rgen.random(size=[n_panels, n_draws, n_mixtures], dtype=np.float32)
            + np.arange(n_draws, dtype=np.float32)[np.newaxis, :, np.newaxis]
        )
        for i in range(n_mixtures):
            for p in range(n_panels):
                rgen.shuffle(draws[p, :, i])
        return np.clip(draws / n_draws, 0, np.float32(1 - 1e-7)), rgen

    # @jitmethod(static_argnums=(0,1), static_argnames=('n_draws', 'n_mixtures'))
    def _make_random_draws_out(self, n_draws, n_mixtures, random_key):
        draws = jax.random.uniform(
            random_key,
            [n_draws, n_mixtures],
        )

        def body(i, carry):
            draws, rnd_key = carry
            rnd_key, subkey = jax.random.split(rnd_key)
            draws = draws.at[:, i].add(jax.random.permutation(subkey, draws.shape[0]))
            return draws, rnd_key

        draws, random_key = jax.lax.fori_loop(
            0, draws.shape[1], body, (draws, random_key)
        )
        return draws * (1 / n_draws), random_key

    # @jitmethod(static_argnums=(0,1,2), static_argnames=('n_draws', 'n_mixtures', 'n_cases'))
    def _make_random_draws_out_2(self, n_draws, n_mixtures, n_cases, random_key):
        def body(carry, x):
            rkey = carry
            draws, rkey = self._make_random_draws_out(n_draws, n_mixtures, rkey)
            return rkey, draws

        random_key, draws = jax.lax.scan(body, random_key, None, n_cases)
        return draws, random_key

    def check_random_draws(self, engine="numpy"):
        if self.mixtures:
            if self.prerolled_draws:
                if self._draws is None:
                    self.make_random_draws(engine=engine)

    def apply_random_draws(self, parameters, draws=None):
        # if draws is None:
        #     draws = self._draws
        parameters = jnp.broadcast_to(
            parameters, [*draws.shape[:-1], parameters.shape[0]]
        )
        for mix_n, mix in enumerate(self.mixtures):
            u = draws[..., mix_n]
            parameters = mix.roll(u, parameters)
        return parameters

    @jitmethod
    def _jax_quantity(self, params, databundle):
        ca = databundle.get("ca", None)
        n_alts = self.dataset.dc.n_alts
        n_nodes = len(self.graph)
        u = jnp.zeros([n_nodes])
        if ca is not None and self._fixed_arrays.qca_param_slot.size:
            u = u.at[:n_alts].add(
                jnp.dot(
                    ca[..., self._fixed_arrays.qca_data_slot],
                    jnp.exp(params[self._fixed_arrays.qca_param_slot])
                    * self._fixed_arrays.qca_scale,
                )
            )
        return u

    @jitmethod
    def jax_quantity(self, params):
        ca = _get_jnp_array(self.dataset, "ca")
        av = _as_jnp_array(self._data_arrays.av)
        if av is not None:
            depth = av.ndim - 1
        elif ca is not None:
            depth = ca.ndim - 2
        else:
            raise ValueError("missing data")
        f = self._jax_quantity
        for _level in range(depth):
            f = jax.vmap(f, in_axes=(None, 0))
        return f(params, {"ca": ca, "av": av})

    @jitmethod
    def _jax_utility(self, params, databundle):
        ca = databundle.get("ca", None)
        co = databundle.get("co", None)
        av = databundle.get("av", None)
        if co is None:
            n_vars_co = 0
        else:
            n_vars_co = co.shape[-1]
        n_alts = self.dataset.dc.n_alts
        n_nodes = len(self.graph)
        x = jnp.zeros([self.dataset.dc.n_alts, n_vars_co + 1])
        x = x.at[self._fixed_arrays.uco_alt_slot, self._fixed_arrays.uco_data_slot].add(
            params[self._fixed_arrays.uco_param_slot] * self._fixed_arrays.uco_scale
        )
        if self._fixed_arrays.qca_param_slot.size:
            theta = params[self._fixed_arrays.qscale_param_slot]
            q = self._jax_quantity(params, databundle)
            log_q = jnp.log(jnp.clip(q, 1e-15, 1e15))
            u = jnp.clip(log_q, -1e15) * theta
        else:
            u = jnp.zeros([n_nodes])
        if ca is not None and self._fixed_arrays.uca_param_slot.size:
            u = u.at[:n_alts].add(
                jnp.dot(
                    ca[..., self._fixed_arrays.uca_data_slot],
                    params[self._fixed_arrays.uca_param_slot]
                    * self._fixed_arrays.uca_scale,
                )
            )
        if co is not None:
            temp = jnp.dot(co, x[:, :-1].T)
            u = u.at[:n_alts].add(temp)
        u = u.at[:n_alts].add(x[:, -1].T)
        if av is not None and not self.availability_any:
            u = u.at[:n_alts].set(jnp.where(av[:n_alts], u[:n_alts], -jnp.inf))
        return u

    @jitmethod
    def jax_utility(self, params):
        ca = _get_jnp_array(self.dataset, "ca")
        co = _get_jnp_array(self.dataset, "co")
        av = _as_jnp_array(self._data_arrays.av)
        if av is not None:
            depth = av.ndim - 1
        elif co is not None:
            depth = co.ndim - 1
        elif ca is not None:
            depth = ca.ndim - 2
        else:
            raise ValueError("missing data")
        f = self._jax_utility
        for _level in range(depth):
            f = jax.vmap(f, in_axes=(None, 0))
        return f(params, {"ca": ca, "co": co, "av": av})

    def __utility_for_nest(self, slot):
        nest_code = self.graph.standard_sort[slot]
        child_slots = self.graph.successor_slots(nest_code)
        mu_name = self.graph.nodes[nest_code].get("parameter")
        if mu_name is None:
            mu_slot = -1
        else:
            mu_slot = self.get_param_loc(mu_name)

        # @jit
        def u_nest(params, utility_array, array_av):
            mu = params[mu_slot] if mu_slot >= 0 else 1.0

            shifter = utility_array[jnp.asarray(child_slots)].max()
            carry = jnp.zeros(utility_array.shape[:-1])
            # num_av = jnp.sum(utility_array[..., child_slots] > -1e30, axis=-1)
            num_av = array_av[..., slot]

            def body(carry, child_slot):
                carry = jnp.add(
                    carry,
                    jnp.where(
                        num_av > 1,
                        jnp.exp(
                            jnp.clip(utility_array[..., child_slot] - shifter, -1e37)
                            / mu
                        ),
                        jnp.exp(
                            jnp.clip(utility_array[..., child_slot] - shifter, -1e37)
                        ),
                    ),
                )
                return carry, None

            carry, _ = jax.lax.scan(body, carry, child_slots)
            utility_array = utility_array.at[..., slot].add(
                jnp.where(
                    num_av > 1,
                    jnp.clip(jnp.log(carry), -1e38) * mu,
                    jnp.clip(jnp.log(carry), -1e39),
                )
                + shifter
            )
            return utility_array

        return u_nest

    def _mu_slots(self):
        n_params = self.n_params
        mu_names = [
            self.graph.nodes[i].get("parameter") for i in self.graph.standard_sort
        ]
        mu_slots = jnp.asarray(
            [
                (self.get_param_loc(mu_name) if mu_name is not None else n_params)
                for mu_name in mu_names
            ]
        )
        return mu_slots

    @compiledmethod
    def utility_for_nests(self):
        n_alts = self.graph.n_elementals()
        n_nodes = len(self.graph)

        def u_nesting_none(out, beta, array_av, mu_slots):
            # TODO Maybe filter on av
            return out.at[..., -1].set(jnp.log(jnp.exp(out[..., :-1]).sum(-1)))

        if n_nodes - n_alts <= 1:
            return u_nesting_none

        if n_nodes - n_alts < 3:
            # @jit
            def u_nesting_few(out, beta, array_av, mu_slots):
                for slot in range(n_alts, n_nodes):
                    out = self.__utility_for_nest(slot)(beta, out, array_av)
                return out

            return u_nesting_few

        # many nests, use more efficient loop
        def add_u_to_parent(u, params, child_slot, parent_slot, avail, mu_slots):
            mu = params[mu_slots[parent_slot]]
            u = u.at[..., parent_slot].add(
                jnp.where(
                    avail[..., parent_slot] > 1,
                    jnp.exp(jnp.clip(u[..., child_slot], -1e37) / mu),
                    jnp.exp(jnp.clip(u[..., child_slot], -85)),
                )
            )
            return u

        def max_u_to_parent(u, bs, child_slot, parent_slot):
            bs = bs.at[..., parent_slot].set(
                jnp.maximum(
                    u[..., child_slot],
                    bs[..., parent_slot],
                )
            )
            return bs

        def log_self(u, bs, params, self_slot, avail, mu_slots):
            mu = params[mu_slots[self_slot]]
            u = u.at[..., self_slot].set(
                jnp.maximum(
                    jnp.where(
                        avail[..., self_slot] > 1,
                        jnp.log(u[..., self_slot]) * mu,
                        jnp.log(u[..., self_slot]),
                    ),
                    bs[..., self_slot],
                )
            )
            return u

        slotarray = np.stack(self.graph.edge_slot_arrays()).T

        # @jax.jit
        def u_rollup(utility_array, parameter_vector, avail_ca, mu_slots):
            n_params = parameter_vector.size
            params = jnp.ones(n_params + 1, dtype=parameter_vector.dtype)
            params = params.at[:n_params].set(parameter_vector)
            # the backstop prevents underflow when mu is too small
            backstop = jnp.full_like(utility_array, -jnp.inf)

            def body(carry, xs):
                u, bs, params = carry
                up_slot, dn_slot, firstvisit, allocslot = xs
                # if firstvisit >= 0 and dn_slot>=n_alts:
                #     u = log_self(u, params, dn_slot)
                u = jax.lax.cond(
                    (firstvisit >= 0) & (dn_slot >= n_alts),
                    lambda u: log_self(u, bs, params, dn_slot, avail_ca, mu_slots),
                    lambda u: u,
                    operand=u,
                )
                u = add_u_to_parent(u, params, dn_slot, up_slot, avail_ca, mu_slots)
                bs = max_u_to_parent(u, bs, dn_slot, up_slot)
                return (u, bs, params), None

            (utility_array, backstop, _ignore_1), _ignore_2 = jax.lax.scan(
                body, (utility_array, backstop, params), slotarray
            )
            # log utility at root
            utility_array = utility_array.at[..., -1].set(
                jnp.clip(jnp.log(utility_array[..., -1]), -1e38)
            )
            return utility_array

        return u_rollup

    def __probability_for_nest(self, slot):
        nest_code = self.graph.standard_sort[slot]
        child_slots = self.graph.successor_slots(nest_code)
        mu_name = self.graph.nodes[nest_code].get("parameter")
        if mu_name is None:
            mu_slot = -1
        else:
            mu_slot = self.get_param_loc(mu_name)

        # @jit
        def probability_nest(params, utility_array, probability_array):
            mu = params[mu_slot] if mu_slot >= 0 else 1.0
            u_nest = utility_array[..., slot]

            def body(carry, child_slot):
                diff = jnp.clip(utility_array[..., child_slot], -1e33) - jnp.clip(
                    u_nest, -1e33
                )
                add_me = diff / mu
                carry = carry.at[..., child_slot].set(add_me + carry[..., slot])
                return carry, None

            probability_array, _ = jax.lax.scan(body, probability_array, child_slots)
            return probability_array

        return probability_nest

    @jitmethod
    def _jax_log_probability_bundle(self, params, databundle):
        ca, co, ch, av, wt = databundle
        return self._jax_log_probability(params, ca, co, av)

    @jitmethod
    def _jax_utility_include_nests(self, params, databundle):
        av = databundle.get("av", None)
        n_alts = self.dataset.dc.n_alts
        utility_array = self._jax_utility(params, databundle)
        # downshift to prevent over/underflow
        shifter = utility_array[:n_alts].max(axis=-1)
        if av is not None and not self.availability_any:
            utility_array = utility_array.at[:n_alts].add(
                jnp.where(av[:n_alts], -shifter, 0)
            )
        else:
            utility_array = utility_array.at[:n_alts].add(-shifter)
        mu_slots = self._mu_slots()
        utility_array = self.utility_for_nests(utility_array, params, av, mu_slots)
        return utility_array + shifter

    @jitmethod
    def jax_utility_include_nests(self, params):
        ca = _get_jnp_array(self.dataset, "ca")
        co = _get_jnp_array(self.dataset, "co")
        av = _as_jnp_array(self._data_arrays.av)
        if av is not None:
            depth = av.ndim - 1
        elif co is not None:
            depth = co.ndim - 1
        elif ca is not None:
            depth = ca.ndim - 2
        else:
            raise ValueError("missing data")
        f = self._jax_utility_include_nests
        for _level in range(depth):
            f = jax.vmap(f, in_axes=(None, 0))
        return f(params, {"ca": ca, "co": co, "av": av})

    def utility(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        return_format=None,
    ):
        if self.compute_engine != "jax":
            return super().utility(
                x=x,
                start_case=start_case,
                stop_case=stop_case,
                step_case=step_case,
                return_format=return_format,
            )
        if x is not None:
            self.pvals = x
        return self.jax_utility(self.pvals)

    @jitmethod
    def _jax_log_probability(self, params, databundle):
        av = databundle.get("av", None)
        n_alts = self.dataset.dc.n_alts
        n_nodes = len(self.graph)
        utility_array = self._jax_utility(params, databundle)
        # downshift to prevent over/underflow
        shifter = utility_array[:n_alts].max(axis=-1)
        if av is not None and not self.availability_any:
            utility_array = utility_array.at[:n_alts].add(
                jnp.where(av[:n_alts], -shifter, 0)
            )
        else:
            utility_array = utility_array.at[:n_alts].add(-shifter)
        mu_slots = self._mu_slots()
        utility_array = self.utility_for_nests(utility_array, params, av, mu_slots)
        logprobability = jnp.zeros_like(utility_array)
        for slot in range(n_nodes, n_alts, -1):
            logprobability = self.__probability_for_nest(slot - 1)(
                params, utility_array, logprobability
            )
        # clipping log probability at 0, prevents underflow/overflow
        # when one of the nests only includes very bad alternatives.
        logprobability = jnp.clip(logprobability, None, 0)
        return logprobability

    @jitmethod
    def jax_log_probability(self, params):
        ca = _get_jnp_array(self.dataset, "ca")
        co = _get_jnp_array(self.dataset, "co")
        av = _as_jnp_array(self._data_arrays.av)
        if av is not None:
            depth = av.ndim - 1
        elif co is not None:
            depth = co.ndim - 1
        elif ca is not None:
            depth = ca.ndim - 2
        else:
            raise ValueError("missing data")
        f = self._jax_log_probability
        for _level in range(depth):
            f = jax.vmap(f, in_axes=(None, 0))
        return f(params, {"ca": ca, "co": co, "av": av})

    @jitmethod
    def _jax_probability(self, params, databundle):
        n_alts = self.dataset.dc.n_alts
        return jnp.exp(self._jax_log_probability(params, databundle)[:n_alts])

    @jitmethod
    def jax_probability(self, params):
        ca = _get_jnp_array(self.dataset, "ca")
        co = _get_jnp_array(self.dataset, "co")
        av = _as_jnp_array(self._data_arrays.av)
        if av is not None:
            depth = av.ndim - 1
        elif co is not None:
            depth = co.ndim - 1
        elif ca is not None:
            depth = ca.ndim - 2
        else:
            raise ValueError("missing data")
        f = self._jax_probability
        for _level in range(depth):
            f = jax.vmap(f, in_axes=(None, 0))
        return f(params, {"ca": ca, "co": co, "av": av})

    @jitmethod
    def _jax_likelihood(self, params, databundle):
        n_alts = self.dataset.dc.n_alts
        ch = databundle.get("ch", None)[:n_alts]
        pr = self._jax_probability(params, databundle)[:n_alts]
        likely = jnp.where(ch, pr, 1.0)  # TODO make power if needed
        return likely

    @jitmethod(static_argnums=(3,), static_argnames="n_draws")
    def _jax_random_params(self, params, databundle, groupbundle=None, n_draws=100):
        if self.prerolled_draws:
            draws = groupbundle.get("draws", None)
        else:
            rk = groupbundle.get("rk", None)
            draws, _ = self._make_random_draws_out(n_draws, len(self.mixtures), rk)
        rand_params = self.apply_random_draws(params, draws)
        return rand_params

    @jitmethod
    def jax_random_params(self, params):
        ca = _get_jnp_array(self.dataset, "ca")
        co = _get_jnp_array(self.dataset, "co")
        av = _as_jnp_array(self._data_arrays.av)
        ch = _as_jnp_array(self._data_arrays.ch)
        n_draws = self.n_draws
        seed = getattr(self, "seed", 42)
        if av is not None:
            depth = av.ndim - 1
            shape = av.shape[:-1]
        elif co is not None:
            depth = co.ndim - 1
            shape = co.shape[:-1]
        elif ca is not None:
            depth = ca.ndim - 2
            shape = ca.shape[:-2]
        elif ch is not None:
            depth = ch.ndim - 1
            shape = ch.shape[:-1]
        else:
            raise ValueError("missing data")
        random_key = jax.random.PRNGKey(seed)
        if self.groupid is not None:
            depth = depth - 1
            shape = shape[:-1]
        f = self._jax_random_params  # params, databundle, groupbundle=None, n_draws=100
        from .random import keysplit

        commons = None if self.common_draws else 0
        for _i in range(depth):
            f = jax.vmap(f, in_axes=(None, 0, commons, None))
            if not self.prerolled_draws:
                random_key, shape = keysplit(random_key, shape)
        if self.prerolled_draws:
            return f(
                params,
                {"ca": ca, "co": co, "av": av, "ch": ch},
                {"draws": self._draws},
                n_draws,
            )
        else:
            return f(
                params,
                {"ca": ca, "co": co, "av": av, "ch": ch},
                {"rk": random_key},
                n_draws,
            )

    @jitmethod(static_argnums=(3,), static_argnames="n_draws")
    def _jax_loglike_casewise(self, params, databundle, groupbundle=None, n_draws=100):
        if len(self.mixtures) == 0:
            logpr = self._jax_log_probability(params, databundle)
            ch = databundle.get("ch", None)
            n_alts = self.dataset.dc.n_alts
            # return (logpr[:n_alts] * ch[:n_alts]).sum()
            return jnp.where(ch[:n_alts], logpr[:n_alts] * ch[:n_alts], 0).sum()
        else:
            if self.prerolled_draws:
                draws = groupbundle.get("draws", None)
            else:
                rk = groupbundle.get("rk", None)
                draws, _ = self._make_random_draws_out(n_draws, len(self.mixtures), rk)
            ch = databundle.get("ch", None)
            rand_params = self.apply_random_draws(params, draws)
            if ch.ndim == 2:  # PANEL DATA
                # vmap over ingroup
                likelihood_f = jax.vmap(
                    self._jax_likelihood,
                    in_axes=(None, 0),
                )
                # vmap over draws
                likelihood = jax.vmap(
                    likelihood_f,
                    in_axes=(0, None),
                    out_axes=-1,
                )(rand_params, databundle)
                # collapse likelihood over all alternatives
                likelihood = likelihood.prod([0, 1])
                # average over all draws
                likelihood = likelihood.mean(0)
                return jnp.log(likelihood)
            else:
                # vmap over draws
                likelihood = jax.vmap(
                    self._jax_likelihood,
                    in_axes=(0, None),
                    out_axes=-1,
                )(rand_params, databundle)
                # collapse likelihood over all alternatives
                likelihood = likelihood.prod(0)
                # average over all draws
                likelihood = likelihood.mean(0)
                return jnp.log(likelihood)

    @jitmethod
    def jax_loglike_casewise(self, params):
        ca = _get_jnp_array(self.dataset, "ca")
        co = _get_jnp_array(self.dataset, "co")
        av = _as_jnp_array(self._data_arrays.av)
        ch = _as_jnp_array(self._data_arrays.ch)
        n_draws = self.n_draws
        seed = getattr(self, "seed", 42)
        if av is not None:
            depth = av.ndim - 1
            shape = av.shape[:-1]
        elif co is not None:
            depth = co.ndim - 1
            shape = co.shape[:-1]
        elif ca is not None:
            depth = ca.ndim - 2
            shape = ca.shape[:-2]
        elif ch is not None:
            depth = ch.ndim - 1
            shape = ch.shape[:-1]
        else:
            raise ValueError("missing data")
        random_key = jax.random.PRNGKey(seed)
        if self.groupid is not None:
            depth = depth - 1
            shape = shape[:-1]
        f = (
            self._jax_loglike_casewise
        )  # params, databundle, groupbundle=None, n_draws=100
        from .random import keysplit

        commons = None if self.common_draws else 0
        for _i in range(depth):
            f = jax.vmap(f, in_axes=(None, 0, commons, None))
            if not self.prerolled_draws:
                random_key, shape = keysplit(random_key, shape)
        if self.prerolled_draws:
            return f(
                params,
                {"ca": ca, "co": co, "av": av, "ch": ch},
                {"draws": self._draws},
                n_draws,
            )
        else:
            return f(
                params,
                {"ca": ca, "co": co, "av": av, "ch": ch},
                {"rk": random_key},
                n_draws,
            )

    @jitmethod
    def jax_loglike(self, params):
        wt = _as_jnp_array(self._data_arrays.wt)
        return (self.jax_loglike_casewise(params) * wt).sum()

    def loglike(
        self, x=None, *, start_case=None, stop_case=None, step_case=None, **kwargs
    ):
        if self.compute_engine != "jax":
            return super().loglike(
                x=x,
                start_case=start_case,
                stop_case=stop_case,
                step_case=step_case,
                **kwargs,
            )
        if start_case is not None:
            raise NotImplementedError("start_case with engine=jax")
        if stop_case is not None:
            raise NotImplementedError("stop_case with engine=jax")
        if step_case is not None:
            raise NotImplementedError("step_case with engine=jax")
        # if kwargs:
        #     raise NotImplementedError(f"{kwargs.popitem()[0]} with engine=jax")
        if x is not None:
            self.pvals = x
        self.check_random_draws()
        result = float(self.jax_loglike(self.pvals))
        if start_case is None and stop_case is None and step_case is None:
            self._check_if_best(result)
        return result

    def d_loglike(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        return_series=False,
        **kwargs,
    ):
        if self.compute_engine != "jax":
            return super().d_loglike(
                x=x,
                start_case=start_case,
                stop_case=stop_case,
                step_case=step_case,
                return_series=return_series,
                **kwargs,
            )
        if start_case is not None:
            raise NotImplementedError("start_case with engine=jax")
        if stop_case is not None:
            raise NotImplementedError("stop_case with engine=jax")
        if step_case is not None:
            raise NotImplementedError("step_case with engine=jax")
        # if kwargs:
        #     raise NotImplementedError(f"{kwargs.popitem()[0]} with engine=jax")
        if x is not None:
            self.pvals = x
        self.check_random_draws()
        result = self.jax_d_loglike(self.pvals) * (self.pholdfast == 0)
        if return_series:
            result = pd.Series(result, index=self.pnames)
        return result

    def loglike_casewise(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        **kwargs,
    ):
        if self.compute_engine != "jax":
            return super().loglike_casewise(
                x=x,
                start_case=start_case,
                stop_case=stop_case,
                step_case=step_case,
                **kwargs,
            )
        if start_case is not None:
            raise NotImplementedError("start_case with engine=jax")
        if stop_case is not None:
            raise NotImplementedError("stop_case with engine=jax")
        if step_case is not None:
            raise NotImplementedError("step_case with engine=jax")
        # if kwargs:
        #     raise NotImplementedError(f"{kwargs.popitem()[0]} with engine=jax")
        if x is not None:
            self.pvals = x
        self.check_random_draws()
        result = np.asarray(self.jax_loglike_casewise(self.pvals))
        return result

    def maximize_loglike(
        self,
        *args,
        **kwargs,
    ):
        """
        Maximize the log likelihood.

        Parameters
        ----------
        method : str, optional
            The optimization method to use.  See scipy.optimize for
            most possibilities, or use 'BHHH'. Defaults to SLSQP if
            there are any constraints or finite parameter bounds,
            otherwise defaults to BHHH.
        quiet : bool, default False
            Whether to suppress the dashboard.
        options : dict, optional
            These options are passed through to the `scipy.optimize.minimize`
            function.
        maxiter : int, optional
            Maximum number of iterations.  This argument is just added to
            `options` for most methods.

        Returns
        -------
        dictx
            A dictionary of results, including final log likelihood,
            elapsed time, and other statistics.  The exact items
            included in output will vary by estimation method.

        """
        if self.compute_engine == "jax":
            self.check_random_draws()
            return self.jax_maximize_loglike(*args, **kwargs)
        else:
            return super().maximize_loglike(*args, **kwargs)

    def loglike_null(self, use_cache=True):
        """
        Compute the log likelihood at null values.

        Set all parameter values to the value indicated in the
        "nullvalue" column of the parameter frame, and compute
        the log likelihood with the currently loaded data.  Note
        that the null value for each parameter may not be zero
        (for example, the default null value for logsum parameters
        in a nested logit model is 1).

        Parameters
        ----------
        use_cache : bool, default True
            Use the cached value if available.  Set to -1 to
            raise an exception if there is no cached value.

        Returns
        -------
        float
        """
        if self.compute_engine != "jax":
            return super().loglike_null(use_cache)
        if self._cached_loglike_null is not None and use_cache:
            return self._cached_loglike_null
        elif use_cache == -1:
            raise ValueError("no cached value")
        else:
            self.check_random_draws()
            self._cached_loglike_null = float(self.jax_loglike(self.pnullvals))
            return self._cached_loglike_null



import importlib
import itertools
import logging
import os
import re
from pathlib import Path

import numpy as np
import pandas as pd
import yaml

from larch import DataFrames, Model, P, X
from larch.log import logger_name
from larch.model.model_group import ModelGroup
from larch.util import Dict

from .general import apply_coefficients, explicit_value_parameters

_logger = logging.getLogger(logger_name)


def generate_alternatives(n_persons):
    """
    Generate a dictionary of CDAP alternatives.

    The keys are the names of the patterns, and
    the values are the alternative code numbers.

    Parameters
    ----------
    n_persons : int

    Returns
    -------
    Dict
    """
    basic_patterns = ["M", "N", "H"]
    alt_names = list(
        "".join(i) for i in itertools.product(basic_patterns, repeat=n_persons)
    )
    alt_codes = np.arange(1, len(alt_names) + 1)
    return Dict(zip(alt_names, alt_codes))


def apply_replacements(expression, prefix, tokens):
    """
    Convert general person terms to specific person terms for the CDAP model.

    Parameters
    ----------
    expression : str
        An expression from the "Expression" column
        of cdap_INDIV_AND_HHSIZE1_SPEC.csv, or similar.
    prefix : str
        A prefix to attach to each token in `expression`.
    tokens : list-like of str
        A list of tokens to edit within an expression.

    Returns
    -------
    expression : str
        The modified expression
    """
    for i in tokens:
        expression = re.sub(rf"\b{i}\b", f"{prefix}_{i}", expression)
    return expression


def cdap_base_utility_by_person(model, n_persons, spec, alts=None, value_tokens=()):
    """
    Build the base utility by person for each pattern.

    Parameters
    ----------
    model : larch.Model
    n_persons : int
    spec : pandas.DataFrame
        The base utility by person spec provided by
        the ActivitySim framework.
    alts : dict, optional
        The keys are the names of the patterns, and
        the values are the alternative code numbers,
        as created by `generate_alternatives`.  If not
        given, the alts are automatically regenerated
        using that function.
    value_tokens : list-like of str, optional
        A list of tokens to edit within an the expressions,
        generally the column names of the provided values
        from the estimation data bundle.  Only used when
        `n_persons` is more than 1.
    """
    if n_persons == 1:
        for i in spec.index:
            if not pd.isna(spec.loc[i, "M"]):
                model.utility_co[1] += X(spec.Expression[i]) * P(spec.loc[i, "M"])
            if not pd.isna(spec.loc[i, "N"]):
                model.utility_co[2] += X(spec.Expression[i]) * P(spec.loc[i, "N"])
            if not pd.isna(spec.loc[i, "H"]):
                model.utility_co[3] += X(spec.Expression[i]) * P(spec.loc[i, "H"])
    else:
        if alts is None:
            alts = generate_alternatives(n_persons)
        person_numbers = range(1, n_persons + 1)
        for pnum in person_numbers:
            for i in spec.index:
                for aname, anum in alts.items():
                    z = pnum - 1
                    if not pd.isna(spec.loc[i, aname[z]]):
                        x = apply_replacements(
                            spec.Expression[i], f"p{pnum}", value_tokens
                        )
                        model.utility_co[anum] += X(x) * P(spec.loc[i, aname[z]])


def interact_pattern(n_persons, select_persons, tag):
    """
    Compile a regex pattern to match CDAP alternatives.

    Parameters
    ----------
    n_persons : int
    select_persons : list-like of int
        The persons to be selected.
    tag : str
        The activity letter, currently one of {M,N,H}.

    Returns
    -------
    re.compile
    """
    pattern = ""
    p = 1
    while len(pattern) < n_persons:
        pattern += tag if p in select_persons else "."
        p += 1
    return re.compile(pattern)


def cdap_interaction_utility(model, n_persons, alts, interaction_coef, coefficients):
    person_numbers = list(range(1, n_persons + 1))

    matcher = re.compile("coef_[HMN]_.*")
    interact_coef_map = {}
    for c in coefficients.index:
        if matcher.search(c):
            c_split = c.split("_")
            for j in c_split[2:]:
                interact_coef_map[(c_split[1], j)] = c
                if all(i == "x" for i in j):  # wildcards also map to empty
                    interact_coef_map[(c_split[1], "")] = c

    for (cardinality, activity), coefs in interaction_coef.groupby(
        ["cardinality", "activity"]
    ):
        _logger.info(
            f"{n_persons} person households, interaction cardinality {cardinality}, activity {activity}"
        )
        if cardinality > n_persons:
            continue
        elif cardinality == n_persons:
            this_aname = activity * n_persons
            this_altnum = alts[this_aname]
            for _rowindex, row in coefs.iterrows():
                expression = "&".join(
                    f"(p{p}_ptype == {t})"
                    for (p, t) in zip(person_numbers, row.interaction_ptypes)
                    if t != "*"
                )
                if expression:
                    if (activity, row.interaction_ptypes) in interact_coef_map:
                        linear_component = X(expression) * P(
                            interact_coef_map[(activity, row.interaction_ptypes)]
                        )
                    else:
                        linear_component = X(expression) * P(row.coefficient)
                else:
                    if (activity, row.interaction_ptypes) in interact_coef_map:
                        linear_component = P(
                            interact_coef_map[(activity, row.interaction_ptypes)]
                        )
                    else:
                        linear_component = P(row.coefficient)
                _logger.debug(
                    f"utility_co[{this_altnum} {this_aname}] += {linear_component}"
                )
                model.utility_co[this_altnum] += linear_component
        elif cardinality < n_persons:
            for combo in itertools.combinations(person_numbers, cardinality):
                pattern = interact_pattern(n_persons, combo, activity)
                for aname, anum in alts.items():
                    if pattern.match(aname):
                        for _rowindex, row in coefs.iterrows():
                            expression = "&".join(
                                f"(p{p}_ptype == {t})"
                                for (p, t) in zip(combo, row.interaction_ptypes)
                                if t != "*"
                            )
                            # interaction terms without ptypes (i.e. with wildcards)
                            # only apply when the household size matches the cardinality
                            if expression != "":
                                if (
                                    activity,
                                    row.interaction_ptypes,
                                ) in interact_coef_map:
                                    linear_component = X(expression) * P(
                                        interact_coef_map[
                                            (activity, row.interaction_ptypes)
                                        ]
                                    )
                                else:
                                    linear_component = X(expression) * P(
                                        row.coefficient
                                    )
                                _logger.debug(
                                    f"utility_co[{anum} {aname}] += {linear_component}"
                                )
                                model.utility_co[anum] += linear_component


def cdap_split_data(households, values):
    if "cdap_rank" not in values:
        raise ValueError("assign cdap_rank to values first")
    # only process the first 5 household members
    values = values[values.cdap_rank <= 5]
    cdap_data = {}
    for hhsize, hhs_part in households.groupby(households.hhsize.clip(1, 5)):
        if hhsize == 1:
            v = pd.merge(values, hhs_part.household_id, on="household_id").set_index(
                "household_id"
            )
        else:
            v = (
                pd.merge(values, hhs_part.household_id, on="household_id")
                .set_index(["household_id", "cdap_rank"])
                .unstack()
            )
            v.columns = [f"p{i[1]}_{i[0]}" for i in v.columns]
            for agglom in ["override_choice", "model_choice"]:
                v[agglom] = v[[f"p{p}_{agglom}" for p in range(1, hhsize + 1)]].sum(1)
        cdap_data[hhsize] = v
    return cdap_data


def cdap_dataframes(households, values):
    data = cdap_split_data(households, values)
    dfs = {}
    for hhsize in data.keys():
        alts = generate_alternatives(hhsize)
        dfs[hhsize] = DataFrames(
            co=data[hhsize],
            alt_names=alts.keys(),
            alt_codes=alts.values(),
            av=1,
            ch=data[hhsize].override_choice.map(alts),
        )
    return dfs


# def _cdap_model(households, values, spec1, interaction_coef, coefficients):
#     cdap_data = cdap_dataframes(households, values)
#     m = {}
#     _logger.info(f"building for model 1")
#     m[1] = Model(dataservice=cdap_data[1])
#     cdap_base_utility_by_person(m[1], n_persons=1, spec=spec1)
#     m[1].choice_any = True
#     m[1].availability_any = True
#
#     # Add cardinality into interaction_coef if not present
#     if 'cardinality' not in interaction_coef:
#         interaction_coef['cardinality'] = interaction_coef['interaction_ptypes'].str.len()
#     for s in [2, 3, 4, 5]:
#         _logger.info(f"building for model {s}")
#         m[s] = Model(dataservice=cdap_data[s])
#         alts = generate_alternatives(s)
#         cdap_base_utility_by_person(m[s], s, spec1, alts, values.columns)
#         cdap_interaction_utility(m[s], s, alts, interaction_coef, coefficients)
#         m[s].choice_any = True
#         m[s].availability_any = True
#
#     result = ModelGroup(m.values())
#     explicit_value_parameters(result)
#     apply_coefficients(coefficients, result)
#     return result


def cdap_data(
    name="cdap",
    edb_directory="output/estimation_data_bundle/{name}/",
    coefficients_file="{name}_coefficients.csv",
    interaction_coeffs_file="{name}_interaction_coefficients.csv",
    households_file="../../final_households.csv",
    persons_file="../../final_persons.csv",
    spec1_file="{name}_INDIV_AND_HHSIZE1_SPEC.csv",
    settings_file="{name}_model_settings.yaml",
    chooser_data_file="{name}_values_combined.csv",
):
    edb_directory = edb_directory.format(name=name)
    if not os.path.exists(edb_directory):
        raise FileNotFoundError(edb_directory)

    def read_csv(filename, **kwargs):
        filename = filename.format(name=name)
        return pd.read_csv(os.path.join(edb_directory, filename), **kwargs)

    def read_yaml(filename, **kwargs):
        filename = filename.format(name=name)
        with open(os.path.join(edb_directory, filename)) as f:
            return yaml.load(f, Loader=yaml.SafeLoader, **kwargs)

    settings = read_yaml(settings_file)

    try:
        hhs = read_csv(households_file)
    except FileNotFoundError:
        hhs = pd.read_csv(households_file)

    try:
        persons = read_csv(persons_file)
    except FileNotFoundError:
        persons = pd.read_csv(persons_file)

    asim_cdap = importlib.import_module("activitysim.abm.models.util.cdap")
    person_rank = asim_cdap.assign_cdap_rank(persons)

    coefficients = read_csv(
        coefficients_file,
        index_col="coefficient_name",
        comment="#",
    )

    interaction_coef = read_csv(
        interaction_coeffs_file,
        dtype={
            "interaction_ptypes": str,
        },
        keep_default_na=False,
        comment="#",
    )

    spec1 = read_csv(spec1_file, comment="#")
    values = read_csv(chooser_data_file, comment="#")
    values["cdap_rank"] = person_rank

    return Dict(
        edb_directory=Path(edb_directory),
        person_data=values,
        spec1=spec1,
        interaction_coef=interaction_coef,
        coefficients=coefficients,
        households=hhs,
        settings=settings,
    )


def cdap_model(
    edb_directory="output/estimation_data_bundle/{name}/",
    coefficients_file="{name}_coefficients.csv",
    interaction_coeffs_file="{name}_interaction_coefficients.csv",
    households_file="../../final_households.csv",
    persons_file="../../final_persons.csv",
    spec1_file="{name}_INDIV_AND_HHSIZE1_SPEC.csv",
    settings_file="{name}_model_settings.yaml",
    chooser_data_file="{name}_values_combined.csv",
    return_data=False,
):
    d = cdap_data(
        name="cdap",
        edb_directory=edb_directory,
        coefficients_file=coefficients_file,
        interaction_coeffs_file=interaction_coeffs_file,
        households_file=households_file,
        persons_file=persons_file,
        spec1_file=spec1_file,
        settings_file=settings_file,
        chooser_data_file=chooser_data_file,
    )

    households = d.households
    values = d.person_data
    spec1 = d.spec1
    interaction_coef = d.interaction_coef
    coefficients = d.coefficients

    cdap_dfs = cdap_dataframes(households, values)
    m = {}
    _logger.info("building for model 1")
    m[1] = Model(dataservice=cdap_dfs[1])
    cdap_base_utility_by_person(m[1], n_persons=1, spec=spec1)
    m[1].choice_any = True
    m[1].availability_any = True

    # Add cardinality into interaction_coef if not present
    if "cardinality" not in interaction_coef:
        interaction_coef["cardinality"] = interaction_coef[
            "interaction_ptypes"
        ].str.len()
    for s in [2, 3, 4, 5]:
        _logger.info(f"building for model {s}")
        m[s] = Model(dataservice=cdap_dfs[s])
        alts = generate_alternatives(s)
        cdap_base_utility_by_person(m[s], s, spec1, alts, values.columns)
        cdap_interaction_utility(m[s], s, alts, interaction_coef, coefficients)
        m[s].choice_any = True
        m[s].availability_any = True

    model = ModelGroup(m.values())
    explicit_value_parameters(model)
    apply_coefficients(coefficients, model)
    if return_data:
        return model, d
    return model



import keyword
import re
import warnings


class NotAPythonIdentifier(Warning):
    pass


def make_valid_identifier(x, suppress_warnings=False):
    x = str(x)
    if keyword.iskeyword(x):
        y = "_" + x
        if not suppress_warnings:
            warnings.warn(
                f"name {x} is a python keyword, converting to {y}",
                stacklevel=2,
            )
    else:
        y = x
    replacer = re.compile(r"(\W+)")
    y = replacer.sub("_", y.strip())
    if not y.isidentifier():
        y = "_" + y
    if y != x:
        if not suppress_warnings:
            warnings.warn(
                f"name {x} is not a valid python identifier, converting to {y}",
                stacklevel=2,
                category=NotAPythonIdentifier,
            )
    return y


def valid_identifier_or_parenthized_string(x, leading_dot=True):
    x = str(x)
    if x.isidentifier():
        if leading_dot:
            return "." + x
        return x
    else:
        if "'" in x and '"' not in x:
            return f'("{x}")'
        if "'" not in x and '"' in x:
            return f"('{x}')"
        raise NotImplementedError("cannot handle strings with both quote types")


def parenthize(x, signs_qualify=False):
    """Wrap a string in parenthesis if needed for unambiguous clarity.

    Parameters
    ----------
    x : str
            The string to wrap
    signs_qualify : bool
            If True, a leading + or - on a number triggers the parenthesis (defaults False)

    """
    x = str(x).strip()
    replacer = re.compile(r"(\W+)")
    if replacer.search(x):
        if signs_qualify:
            numeric = re.compile(
                r"^(([1-9][0-9]*\.?[0-9]*)|(\.[0-9]+))([Ee][+-]?[0-9]+)?\Z"
            )
        else:
            numeric = re.compile(
                r"^[+-]?(([1-9][0-9]*\.?[0-9]*)|(\.[0-9]+))([Ee][+-]?[0-9]+)?\Z"
            )
        if numeric.search(x):
            return x
        return f"({x})"
    return x



def example(extract='m', estimate=False):
    import numpy as np
    import pandas as pd

    import larch as lx
    from larch import P, X

    hh, pp, tour, skims, emp = lx.example(200, ["hh", "pp", "tour", "skims", "emp"])

    hh["INCOME_GRP"] = pd.qcut(hh.INCOME, 3)

    co = lx.Dataset.construct(
        tour.set_index("TOURID"),
        caseid="TOURID",
        alts=skims.TAZ_ID,
    )
    co

    emp.info()

    tree = lx.DataTree(
        base=co,
        hh=hh.set_index("HHID"),
        person=pp.set_index("PERSONID"),
        emp=emp,
        skims=lx.Dataset.construct.from_omx(skims),
        relationships=(
            "base.TAZ_ID @ emp.TAZ",
            "base.HHID @ hh.HHID",
            "base.PERSONID @ person.PERSONID",
            "hh.HOMETAZ @ skims.otaz",
            "base.TAZ_ID @ skims.dtaz",
        ),
    ).digitize_relationships()

    m = lx.Model(datatree=tree)
    m.title = "Exampville Tour Destination Choice v2"

    m.quantity_ca = (
        +P.EmpRetail_HighInc * X("RETAIL_EMP * (INCOME>50000)")
        + P.EmpNonRetail_HighInc * X("NONRETAIL_EMP") * X("INCOME>50000")
        + P.EmpRetail_LowInc * X("RETAIL_EMP") * X("INCOME<=50000")
        + P.EmpNonRetail_LowInc * X("NONRETAIL_EMP") * X("INCOME<=50000")
    )

    m.quantity_scale = P.Theta

    m.utility_ca = +P.distance * X.AUTO_DIST

    m.choice_co_code = "base.DTAZ"

    m.plock(EmpRetail_HighInc=0, EmpRetail_LowInc=0)

    mj = m.copy()

    m.compute_engine = "numba"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True)

    resultj = mj.maximize_loglike(stderr=True)

    resultj



    m.histogram_on_idca_variable("AUTO_DIST")

    m.histogram_on_idca_variable("RETAIL_EMP")

    m.histogram_on_idca_variable("AUTO_DIST", bins=40, span=(0, 10))

    m.histogram_on_idca_variable(
        "AUTO_DIST",
        x_label="Distance (miles)",
        bins=26,
        span=(0, 13),
        filter_co="INCOME<10000",
    )



    tour_plus = tour.join(hh.set_index("HHID")[["HOMETAZ", "INCOME"]], on="HHID")
    tour_plus["LOW_INCOME"] = tour_plus.INCOME < 50_000
    tour_agg = (
        tour_plus.groupby(["HOMETAZ", "DTAZ", "LOW_INCOME"])
        .size()
        .unstack("DTAZ")
        .fillna(0)
    )

    # j = tour_agg.reset_index(drop=True)
    # lx.DataArray(j.values, dims=("index", "DTAZ"), coords={"index": j.index, "DTAZ": j.columns})

    agg_dataset = lx.Dataset.construct.from_idco(
        tour_agg.index.to_frame().reset_index(drop=True)
    )
    j = tour_agg.reset_index(drop=True)
    agg_dataset = agg_dataset.assign(
        destinations=lx.DataArray(
            j.values,
            dims=("index", "DTAZ"),
            coords={"index": j.index, "DTAZ": j.columns},
        )
    )
    agg_dataset.dc.ALTID = "DTAZ"
    agg_dataset

    agg_tree = lx.DataTree(
        base=agg_dataset,
        emp=emp,
        skims=lx.Dataset.construct.from_omx(skims),
        relationships=(
            "base.DTAZ @ emp.TAZ",
            "base.HOMETAZ @ skims.otaz",
            "base.DTAZ @ skims.dtaz",
        ),
    )

    mg = lx.Model(datatree=agg_tree, compute_engine="numba")
    mg.title = "Exampville Semi-Aggregate Destination Choice"

    mg.quantity_ca = (
        +P.EmpRetail_HighInc * X("RETAIL_EMP * (1-LOW_INCOME)")
        + P.EmpNonRetail_HighInc * X("NONRETAIL_EMP") * X("(1-LOW_INCOME)")
        + P.EmpRetail_LowInc * X("RETAIL_EMP") * X("LOW_INCOME")
        + P.EmpNonRetail_LowInc * X("NONRETAIL_EMP") * X("LOW_INCOME")
    )

    mg.quantity_scale = P.Theta

    mg.utility_ca = +P.distance * X.AUTO_DIST

    mg.choice_ca_var = "base.destinations"

    mg.plock(EmpRetail_HighInc=0, EmpRetail_LowInc=0)

    result = mg.maximize_loglike(stderr=True)
    result
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



import logging
import os
from pathlib import Path

import pandas as pd
import yaml

from larch import DataFrames, Model
from larch.util import Dict

from ...log import logger_name
from .general import (
    apply_coefficients,
    cv_to_ca,
    linear_utility_from_spec,
    remove_apostrophes,
)

_logger = logging.getLogger(logger_name)


def interaction_simulate_data(
    name="non_mandatory_tour_frequency",
    edb_directory="output/estimation_data_bundle/{name}/",
    settings_file="{name}_model_settings.yaml",
    spec_file="{name}_SPEC.csv",
    alt_def_file="{name}_alternatives.csv",
    coefficients_files="{segment_name}/{name}_coefficients_{segment_name}.csv",
    chooser_data_files="{segment_name}/{name}_choosers_combined.csv",
    alt_values_files="{segment_name}/{name}_interaction_expression_values.csv",
):
    edb_directory = edb_directory.format(name=name)

    def _read_csv(filename, **kwargs):
        filename = filename.format(name=name)
        return pd.read_csv(os.path.join(edb_directory, filename), **kwargs)

    settings_file = settings_file.format(name=name)
    with open(os.path.join(edb_directory, settings_file)) as yf:
        settings = yaml.load(
            yf,
            Loader=yaml.SafeLoader,
        )

    coefficients = {}
    chooser_data = {}
    alt_values = {}

    segment_names = [s["NAME"] for s in settings["SPEC_SEGMENTS"]]

    for segment_name in segment_names:
        coefficients[segment_name] = _read_csv(
            coefficients_files.format(name=name, segment_name=segment_name),
            index_col="coefficient_name",
        )
        chooser_data[segment_name] = _read_csv(
            chooser_data_files.format(name=name, segment_name=segment_name),
        )
        alt_values[segment_name] = _read_csv(
            alt_values_files.format(name=name, segment_name=segment_name),
        )

    spec = _read_csv(
        spec_file,
    )
    spec = remove_apostrophes(spec, ["Label"])
    # alt_names = list(spec.columns[3:])
    # alt_codes = np.arange(1, len(alt_names) + 1)
    # alt_names_to_codes = dict(zip(alt_names, alt_codes))
    # alt_codes_to_names = dict(zip(alt_codes, alt_names))

    alt_def = _read_csv(alt_def_file.format(name=name), index_col=0)

    return Dict(
        edb_directory=Path(edb_directory),
        settings=settings,
        chooser_data=chooser_data,
        coefficients=coefficients,
        alt_values=alt_values,
        spec=spec,
        alt_def=alt_def,
    )


def link_same_value_coefficients(segment_names, coefficients, spec):
    # Assume all coefficients with exactly equal current values are
    # actually the same estimated coefficient value and should be
    # treated as such by Larch.  Comment out this function where called to relax
    # this assumption, although be careful about the number of unique
    # parameters to estimate in these models.

    relabel_coef = {}
    for segment_name in segment_names:
        coef_backwards_map = dict(
            [(j, i) for i, j in coefficients[segment_name]["value"].items()]
        )
        relabel_coef[segment_name] = r = coefficients[segment_name]["value"].map(
            coef_backwards_map
        )
        spec[segment_name] = spec[segment_name].map(r)


def unavail_parameters(model):
    return model.pf.index[(model.pf.value < -900) & (model.pf.holdfast != 0)]


def unavail_data_cols(model):
    locks = unavail_parameters(model)
    return [i.data for i in model.utility_ca if i.param in locks]


def unavail(model, x_ca):
    lock_data = unavail_data_cols(model)
    if len(lock_data):
        unav = x_ca[lock_data[0]] > 0
        for j in lock_data[1:]:
            unav |= x_ca[j] > 0
    return unav


def nonmand_tour_freq_model(
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    data = interaction_simulate_data(
        name="non_mandatory_tour_frequency",
        edb_directory=edb_directory,
    )

    settings = data.settings
    segment_names = [s["NAME"] for s in settings["SPEC_SEGMENTS"]]
    link_same_value_coefficients(segment_names, data.coefficients, data.spec)
    spec = data.spec
    coefficients = data.coefficients
    chooser_data = data.chooser_data
    alt_values = data.alt_values
    alt_def = data.alt_def

    m = {}
    for segment_name in segment_names:
        segment_model = m[segment_name] = Model()
        # One of the alternatives is coded as 0, so
        # we need to explicitly initialize the MNL nesting graph
        # and set to root_id to a value other than zero.
        segment_model.initialize_graph(alternative_codes=alt_def.index, root_id=9999)

        ### Utility specifications
        segment_model.utility_ca = linear_utility_from_spec(
            spec,
            x_col="Label",
            p_col=segment_name,
        )
        apply_coefficients(coefficients[segment_name], segment_model)
        segment_model.choice_co_code = "override_choice"

        ### Attach Data
        x_co = (
            chooser_data[segment_name]
            .set_index("person_id")
            .rename(columns={"TAZ": "HOMETAZ"})
        )
        x_ca = cv_to_ca(alt_values[segment_name].set_index(["person_id", "variable"]))
        d = DataFrames(
            co=x_co,
            ca=x_ca,
            av=~unavail(segment_model, x_ca),
        )
        m[segment_name].dataservice = d

    if return_data:
        return m, data
    return m



import numpy as np
import pandas as pd
import pandas.core.algorithms as algos


def remove_unused_level(multiindex, level=0):
    """
    Remove unused labels from a level.

    Create a new MultiIndex from the current that removes
    unused levels from a given level[s], meaning that they are
    not expressed in that level's labels.
    The resulting MultiIndex will have the same outward
    appearance, meaning the same .values and ordering. It will also
    be .equals() to the original.

    Parameters
    ----------
    multiindex : pandas.MultiIndex
    level : int or List[int]

    Returns
    -------
    pandas.MultiIndex

    """
    new_levels = []
    new_codes = []
    if isinstance(level, int):
        level = (level,)

    changed = False
    for n, (lev, level_codes) in enumerate(zip(multiindex.levels, multiindex.codes)):
        if n in level:
            # Since few levels are typically unused, bincount() is more
            # efficient than unique() - however it only accepts positive values
            # (and drops order):
            uniques = np.where(np.bincount(level_codes + 1) > 0)[0] - 1
            has_na = int(len(uniques) and (uniques[0] == -1))

            if len(uniques) != len(lev) + has_na:
                # We have unused levels
                changed = True

                # Recalculate uniques, now preserving order.
                # Can easily be cythonized by exploiting the already existing
                # "uniques" and stop parsing "level_codes" when all items
                # are found:
                uniques = algos.unique(level_codes)
                if has_na:
                    na_idx = np.where(uniques == -1)[0]
                    # Just ensure that -1 is in first position:
                    uniques[[0, na_idx[0]]] = uniques[[na_idx[0], 0]]

                # codes get mapped from uniques to 0:len(uniques)
                # -1 (if present) is mapped to last position
                code_mapping = np.zeros(len(lev) + has_na)
                # ... and reassigned value -1:
                code_mapping[uniques] = np.arange(len(uniques)) - has_na

                level_codes = code_mapping[level_codes]

                # new levels are simple
                lev = lev.take(uniques[has_na:])

        new_levels.append(lev)
        new_codes.append(level_codes)

    result = multiindex.view()

    if changed:
        result._reset_identity()
        result._set_levels(new_levels, validate=False)
        result._set_codes(new_codes, validate=False)

    return result


def replace_levels(multiindex, level, new_label_array):
    """
    Replace the labels in a level with a new order.

    Parameters
    ----------
    multiindex : pandas.MultiIndex
    level : int
    new_label_array : array-like

    Returns
    -------
    pandas.MultiIndex
    """
    levels = [i for i in multiindex.levels]
    codes = [i for i in multiindex.codes]

    new_label_array = np.asarray(new_label_array)
    old_label_array = multiindex.levels[level]
    where_in_new_label_array = {}
    for n, j in enumerate(old_label_array):
        try:
            where_in_new_label_array[n] = int(np.where(j == new_label_array)[0])
        except TypeError:
            raise ValueError(
                f"missing {j} in new_label_array, all the old labels must appear"
            ) from None

    old_codes = list(multiindex.codes[level])
    codes[level] = list(map(where_in_new_label_array.get, old_codes))
    levels[level] = new_label_array

    return pd.MultiIndex(levels=levels, codes=codes, names=multiindex.names)



import numpy as np
from numba import njit, prange


@njit(parallel=True, nogil=True)
def cascade_or(arr, dn_slots, up_slots):
    for i in prange(arr.shape[0]):
        for j in range(dn_slots.size):
            arr[i, up_slots[j]] |= arr[i, dn_slots[j]]


@njit(parallel=True, nogil=True)
def cascade_sum(arr, dn_slots, up_slots, max_val=None):
    for i in prange(arr.shape[0]):
        for j in range(dn_slots.size):
            if max_val is None:
                arr[i, up_slots[j]] += arr[i, dn_slots[j]]
            else:
                k = arr[i, up_slots[j]] + arr[i, dn_slots[j]]
                arr[i, up_slots[j]] = min(k, max_val)


def data_av_cascade(dataframes, graph):
    """
    Create an extra wide dataframe with availability rolled up to nests.

    This has all the same elemental alternatives as the original
    `data_av` dataframe, plus columns for all nests.  For each
    nest, if any component is available, then the nest is also
    indicated as available.

    Parameters
    ----------
    dataframes : DataFrames
    graph : NestingTree

    Returns
    -------
    array
    """
    result = np.zeros((len(dataframes.data_av.index), len(graph)), dtype=np.int8)
    result[:, : graph.n_elementals()] = dataframes.data_av
    ups, dns, _1, _2 = graph.edge_slot_arrays()
    cascade_or(result, dns, ups)
    return result


def data_ch_cascade(dataframes, graph, dtype=None):
    """
    Create an extra wide dataframe with choices rolled up to nests.

    This has all the same elemental alternatives as the original
    `data_ch` dataframe, plus columns for all nests.  For each
    nest, if any component is chosen, then the nest is also
    indicated as chosen, with a magnitude equal to the sum of its
    parts.

    Parameters
    ----------
    dataframes : DataFrames
    graph : NestingTree

    Returns
    -------
    array
    """
    result = np.zeros(
        (len(dataframes.data_ch.index), len(graph)),
        dtype=dtype or dataframes.data_ch.dtype,
    )
    result[:, : graph.n_elementals()] = dataframes.data_ch
    ups, dns, _1, _2 = graph.edge_slot_arrays()
    cascade_sum(result, dns, ups)
    return result


def array_av_cascade(arr_av, graph):
    """
    Create an extra wide array with availability rolled up to nests.

    This has all the same elemental alternatives as the original
    `arr_av` array, plus columns for all nests.  For each
    nest, if any component is available, then the nest is also
    indicated as available.

    Parameters
    ----------
    arr_av : ndarray
    graph : NestingTree

    Returns
    -------
    array
    """
    if arr_av is None:
        return None
    result = np.zeros((*arr_av.shape[:-1], len(graph)), dtype=np.int8)
    result[..., : graph.n_elementals()] = arr_av
    ups, dns, _1, _2 = graph.edge_slot_arrays()
    cascade_sum(result.reshape(-1, len(graph)), dns, ups, 127)
    return result


def array_ch_cascade(arr_ch, graph, dtype=None):
    """
    Create an extra wide array with choices rolled up to nests.

    This has all the same elemental alternatives as the original
    `arr_ch` array, plus columns for all nests.  For each
    nest, if any component is chosen, then the nest is also
    indicated as chosen, with a magnitude equal to the sum of its
    parts.

    Parameters
    ----------
    arr_ch : ndarray
    graph : NestingTree

    Returns
    -------
    array
    """
    if arr_ch is None:
        return None
    result = np.zeros(
        (*arr_ch.shape[:-1], len(graph)),
        dtype=dtype or arr_ch.dtype,
    )
    result[..., : graph.n_elementals()] = arr_ch
    ups, dns, _1, _2 = graph.edge_slot_arrays()
    cascade_sum(result.reshape(-1, len(graph)), dns, ups)
    return result
